import logging
import sqlite3


logger = logging.getLogger(__name__)


def make_table(db_file, table, row, primary_key=None):
    """Create a table in the database based on the types in row.

    Parameters
    ----------
    db_file : Path
        Database file. Create if it doesn't already exist.

    table : str
    row : dict
        Each key will be a column in the table. Define schema by the types of the values.
    primary_key : str | None
        Column name to define as the primary key

    """
    schema = []
    type_map = {int: "INTEGER", float: "REAL", str: "TEXT"}
    for name, val in row.items():
        column_type = type_map.get(type(val), "TEXT")
        entry = f"{name} {column_type}"
        if name == primary_key:
            entry += " PRIMARY KEY"
        schema.append(entry)

    con = sqlite3.connect(db_file)
    cur = con.cursor()
    schema_text = ", ".join(schema)
    cur.execute(f"CREATE TABLE {table}({schema_text})")
    con.commit()
    logger.info("Created table=%s in db_file=%s", table, db_file)


def insert_rows(db_file, table, rows):
    """Insert a list of rows into the database table.

    Parameters
    ----------
    db_file : Path
    table : str
    rows : list[tuple]
        Each row should be a tuple of values.

    """
    con = sqlite3.connect(db_file)
    cur = con.cursor()
    placeholder = ""
    num_columns = len(rows[0])
    for i in range(num_columns):
        if i == num_columns - 1:
            placeholder += "?"
        else:
            placeholder += "?, "
    query = f"INSERT INTO {table} VALUES({placeholder})"
    cur.executemany(query, rows)
    con.commit()
    logger.info("Inserted rows into table=%s in db_file=%s", table, db_file)



"""Manages multi-node job coordination"""

import logging
import os
import re
import shutil
import socket
import time
from pathlib import Path

import click
import psutil

from jade.common import CONFIG_FILE, JOBS_OUTPUT_DIR
from jade.extensions.generic_command.generic_command_parameters import GenericCommandParameters
from jade.jobs.job_configuration_factory import create_config_from_file
from jade.loggers import setup_logging
from jade.models.spark import SparkConfigModel, SparkContainerModel
from jade.spark.metrics import SparkMetrics
from jade.utils.run_command import check_run_command, run_command
from jade.utils.utils import get_cli_string, create_script


KiB = 1024
MiB = KiB * KiB
GiB = MiB * KiB
SHUTDOWN_FILENAME = "shutdown"
TMPFS_MOUNT = "/dev/shm"

logger = logging.getLogger(__name__)


@click.command(
    context_settings=dict(
        allow_extra_args=True,
        ignore_unknown_options=True,
    )
)
@click.argument("job_name")
@click.option(
    "-o",
    "--jade-runtime-output",
    required=True,
    type=click.Path(exists=True),
    help="output directory passed to 'jade submit-jobs'",
)
@click.option(
    "--verbose", is_flag=True, default=False, show_default=True, help="Enable verbose log output."
)
@click.argument("manager_script_and_args", nargs=-1, type=click.UNPROCESSED)
def run_spark_cluster(job_name, jade_runtime_output, verbose, manager_script_and_args):
    """Create a Spark cluster across multiple nodes. The manager node will invoke the script."""
    config = create_config_from_file(Path(jade_runtime_output) / CONFIG_FILE)
    job = config.get_job(job_name)
    _set_hostnames(jade_runtime_output)
    output = {}
    check_run_command(f"jade cluster am-i-manager {jade_runtime_output}", output)
    result = output["stdout"].strip()
    manager_node = _get_manager_node_name(jade_runtime_output)
    if result == "true":
        ret = run_cluster_master(
            job, manager_node, jade_runtime_output, verbose, manager_script_and_args
        )
    else:
        assert result == "false", result
        ret = run_worker(job, manager_node, jade_runtime_output, verbose)

    return ret


@click.command()
@click.argument("container")
@click.argument("spark-conf")
@click.argument("output-dir")
@click.option(
    "-s", "--script", help="Optional script to run on cluster. If None, sleep indefinitely."
)
def start_spark_cluster_master(container, spark_conf, output_dir, script):
    job = _create_job(container, spark_conf, script)
    manager_script_and_args = job.model.command.split(" ")
    run_cluster_master(job, socket.gethostname(), output_dir, False, manager_script_and_args)


@click.command()
@click.argument("cluster-master-node-name")
@click.argument("container")
@click.argument("spark-conf")
@click.argument("output-dir")
def start_spark_cluster_worker(cluster_master_node_name, container, spark_conf, output_dir):
    job = _create_job(container, spark_conf, None)
    run_worker(job, cluster_master_node_name, output_dir, False)


def _create_job(container, spark_conf, script):
    if script is None:
        script = Path("/tmp/sleep.sh")
        # TODO: Can we get the walltime value?
        create_script(script, "sleep 10d\n", executable=True)
    return GenericCommandParameters(
        name="1",
        command=f"bash {script}",
        spark_config=SparkConfigModel(
            conf_dir=spark_conf,
            container=SparkContainerModel(path=container),
            enabled=True,
            run_user_script_inside_container=False,
        ),
    )


def run_cluster_master(job, manager_node, output_dir, verbose, manager_script_and_args):
    """Run the cluster master instance."""
    shutdown_file = _get_shutdown_file(job.name, output_dir)
    if shutdown_file.exists():
        os.remove(shutdown_file)
    try:
        return _run_cluster_master(job, manager_node, output_dir, verbose, manager_script_and_args)
    finally:
        # Notify the workers to shutdown.
        shutdown_file.touch()


def _run_cluster_master(job, manager_node, output_dir, verbose, manager_script_and_args):
    filename = os.path.join(output_dir, f"run_spark_cluster__{job.name}.log")
    level = logging.DEBUG if verbose else logging.INFO
    setup_logging(__name__, filename, file_level=level, console_level=level, mode="w")
    logger.info(
        "Run cluster master on %s job=%s: %s", socket.gethostname(), job.name, get_cli_string()
    )

    job_output = Path(output_dir) / JOBS_OUTPUT_DIR / job.name
    if job_output.exists():
        shutil.rmtree(job_output)
    job_output.mkdir(parents=True)
    events_dir = job_output / "spark" / "events"
    events_dir.mkdir(parents=True)
    logs_dir = job_output / "spark" / "logs"
    logs_dir.mkdir()
    workers_dir = job_output / "spark" / "workers"
    workers_dir.mkdir()

    # Make a job-specific conf directory because the log and event files need to be per-job.
    job_conf_dir = job_output / "spark" / "conf"
    shutil.copytree(Path(job.model.spark_config.conf_dir) / "conf", job_conf_dir)
    _fix_spark_conf_file(job_conf_dir, events_dir)
    _set_env_variables(job, job_conf_dir, logs_dir)

    # Ignore errors. Spark may not be running.
    run_command(job.model.spark_config.get_stop_worker())
    run_command(job.model.spark_config.get_stop_history_server())
    run_command(job.model.spark_config.get_stop_master())

    # It would be better to start all workers from the master. Doing so would require that
    # Spark processes on the master node be able to ssh into the worker nodes.
    # I haven't spent the time to figure out to do that inside Singularity containers.
    master_cmd = job.model.spark_config.get_start_master()
    logger.info("Run spark master: [%s]", master_cmd)
    check_run_command(master_cmd)
    history_cmd = job.model.spark_config.get_start_history_server()
    logger.info("Run spark history server: [%s]", history_cmd)
    check_run_command(history_cmd)
    worker_memory = _get_worker_memory_str(job, is_master=True)

    worker_cmd = _get_worker_command(job, manager_node, memory=worker_memory)
    logger.info("Run spark worker: [%s]", worker_cmd)
    check_run_command(worker_cmd)

    # Wait for workers.
    # TODO: find a way to check programmatically with the rest api
    # or parse the logs
    time.sleep(15)
    args = list(manager_script_and_args) + [_get_cluster(manager_node), str(job_output)]
    if job.model.spark_config.run_user_script_inside_container:
        user_cmd = str(job.model.spark_config.get_run_user_script()) + " " + " ".join(args)
    else:
        user_cmd = " ".join(args)
    logger.info("Run user script [%s]", user_cmd)

    start = time.time()
    ret = run_command(user_cmd)
    logger.info("Finished job. duration = %s seconds", time.time() - start)

    # Delay to ensure the history is saved.
    time.sleep(10)
    metrics = SparkMetrics("localhost", history=True)
    try:
        metrics.generate_metrics(job_output / "spark_metrics")
    except Exception:
        logger.exception("Failed to generate metrics")

    check_run_command(job.model.spark_config.get_stop_worker())
    check_run_command(job.model.spark_config.get_stop_history_server())
    check_run_command(job.model.spark_config.get_stop_master())
    if job.model.spark_config.collect_worker_logs:
        shutil.copytree(Path(os.environ["SPARK_WORKER_DIR"]), workers_dir / socket.gethostname())
    return ret


def run_worker(job, manager_node, output_dir, verbose, poll_interval=60):
    """Run a worker instance."""
    logger.error("in worker manager_node=%s job=%s", manager_node, job.name)
    hostname = socket.gethostname()
    filename = os.path.join(output_dir, f"run_spark_job_worker__{hostname}__{job.name}.log")
    level = logging.DEBUG if verbose else logging.INFO
    setup_logging(__name__, filename, file_level=level, console_level=level, mode="w")
    logger.info("Run worker: %s", get_cli_string())

    # Ignore errors. Spark may not be running.
    run_command(job.model.spark_config.get_stop_worker())

    # Give the master a head start.
    time.sleep(10)
    job_output = Path(output_dir) / JOBS_OUTPUT_DIR / job.name
    logs_dir = job_output / "spark" / "logs"
    job_conf_dir = job_output / "spark" / "conf"
    workers_dir = job_output / "spark" / "workers"
    _set_env_variables(job, job_conf_dir, logs_dir)
    worker_memory = _get_worker_memory_str(job, is_master=False)
    cmd = _get_worker_command(job, manager_node, worker_memory)
    ret = 1
    output = {}
    for _ in range(5):
        output.clear()
        logger.info("Run spark worker: [%s]", cmd)
        ret = run_command(cmd, output=output)
        if ret == 0:
            break
    if ret != 0:
        logger.error("Failed to start spark worker: %s: %s", ret, output)

    shutdown_file = _get_shutdown_file(job.name, output_dir)
    while not shutdown_file.exists():
        logger.debug("sleep for %s seconds", poll_interval)
        time.sleep(poll_interval)

    logger.info("Detected shutdown.")
    check_run_command(job.model.spark_config.get_stop_worker())
    if job.model.spark_config.collect_worker_logs:
        shutil.copytree(Path(os.environ["SPARK_WORKER_DIR"]), workers_dir / hostname)
    return 0


def _set_env_variables(job, conf_dir, logs_dir):
    os.environ["SPARK_CONF_DIR"] = str(conf_dir.absolute())
    os.environ["SPARK_LOG_DIR"] = str(logs_dir.absolute())
    if job.model.spark_config.use_tmpfs_for_scratch:
        scratch = TMPFS_MOUNT
    elif job.model.spark_config.alt_scratch is not None:
        scratch = job.model.spark_config.alt_scratch
        Path(scratch).mkdir(parents=True, exist_ok=True)
    else:
        scratch = os.environ["LOCAL_SCRATCH"]
    os.environ["SPARK_LOCAL_DIRS"] = f"{scratch}/spark/local"
    os.environ["SPARK_WORKER_DIR"] = f"{scratch}/spark/worker"


def _get_cluster(manager_node):
    return f"spark://{manager_node}:7077"


def _get_worker_command(job, manager_node, memory):
    cluster = _get_cluster(manager_node)
    return job.spark_config.get_start_worker(memory, cluster)


def _get_manager_node_name(output_dir):
    output = {}
    job_id = os.environ["SLURM_JOB_ID"]  # TODO: needs to be agnostic to HPC type
    check_run_command(f"jade cluster manager-node {output_dir} {job_id}", output)
    return output["stdout"].strip()


def _set_hostnames(output_dir):
    output = {}
    job_id = os.environ["SLURM_JOB_ID"]  # TODO: needs to be agnostic to HPC type
    check_run_command(f"jade cluster hostnames -j {job_id} {output_dir}", output)
    hostnames = [x for x in output["stdout"].split() if x != ""]
    logger.info("Found %s hostnames: %s", len(hostnames), hostnames)
    os.environ["JADE_OUTPUT_DIR"] = output_dir
    os.environ["JADE_COMPUTE_NODE_NAMES"] = " ".join(hostnames)
    return hostnames


def _get_shutdown_file(job_name, output_dir):
    return Path(output_dir) / (SHUTDOWN_FILENAME + "__" + job_name)


def _fix_spark_conf_file(job_conf_dir, events_dir):
    events_dir = events_dir.absolute()
    conf_file = job_conf_dir / "spark-defaults.conf"
    with open(conf_file, "a") as f_out:
        f_out.write(f"spark.eventLog.dir file://{events_dir}\n")
        f_out.write(f"spark.history.fs.logDirectory file://{events_dir}\n")


def _get_tmpfs_size_gb():
    output = {}
    check_run_command("df -h", output=output)
    # Output looks like this:
    # Filesystem                                  Size  Used Avail Use% Mounted on
    # tmpfs                                       378G  4.0K  378G   1% /dev/shm
    for line in output["stdout"].splitlines():
        if line.endswith(TMPFS_MOUNT):
            return _parse_tmpfs_size_str(line)
    raise Exception(f"Did not find {TMPFS_MOUNT} in 'df -h' output: {output['stdout']}")


def _parse_tmpfs_size_str(text):
    fields = text.strip().split()
    if len(fields) < 5:
        raise Exception(f"Output of 'df -h' not understood: {text}")
    avail = fields[3]
    match = re.search(r"(\d+)([MGT])$", avail)
    if match is None:
        raise Exception(f"format of size not understood: {avail}")
    val = int(match.group(1))
    unit = match.group(2)
    if unit == "M":
        val /= KiB
    elif unit == "T":
        val *= KiB
    return val


def _get_total_memory_gib():
    return psutil.virtual_memory()._asdict()["total"] // GiB


def _get_worker_memory_str(job, is_master):
    overhead = job.model.spark_config.node_memory_overhead_gb
    if is_master:
        overhead += job.model.spark_config.master_node_memory_overhead_gb

    if job.model.spark_config.worker_memory_gb == 0:
        total_memory = _get_total_memory_gib()
    else:
        total_memory = job.model.spark_config.worker_memory_gb

    if job.model.spark_config.use_tmpfs_for_scratch:
        total_memory -= _get_tmpfs_size_gb()

    return str(total_memory - overhead) + "g"



"""CLI to try to submit new jobs for an existing submission."""

import logging
import os
import sys

import click

from jade.enums import Status
from jade.jobs.cluster import Cluster
from jade.jobs.job_submitter import JobSubmitter
from jade.loggers import setup_logging, setup_event_logging
from jade.utils.utils import get_cli_string


@click.command()
@click.argument(
    "output",
    type=click.Path(exists=True),
)
@click.option(
    "--verbose", is_flag=True, default=False, show_default=True, help="Enable verbose log output."
)
def try_submit_jobs(output, verbose):
    """Internal command to try to submit new jobs for an existing submission."""
    cluster, promoted = Cluster.deserialize(
        output,
        try_promote_to_submitter=True,
        deserialize_jobs=True,
    )
    if not promoted:
        print("Another node is already the submitter.")
        sys.exit(0)

    # Only create the logger if we get promoted.
    filename = os.path.join(output, "submit_jobs.log")
    event_filename = os.path.join(output, "submit_jobs_events.log")
    setup_event_logging(event_filename, mode="a")
    level = logging.DEBUG if verbose else logging.INFO
    logger = setup_logging(__name__, filename, file_level=level, console_level=level, mode="a")
    logger.info(get_cli_string())

    if cluster.is_complete():
        cluster.demote_from_submitter()
        logger.info("All jobs are already finished.")
        sys.exit(0)

    ret = 1
    try:
        mgr = JobSubmitter.load(output)
        status = mgr.submit_jobs(cluster)
        if status == Status.IN_PROGRESS:
            check_cmd = f"jade show-status -o {output}"
            print(f"Jobs are in progress. Run '{check_cmd}' for updates.")
            ret = 0
        else:
            ret = status.value
    except Exception:
        logger.exception("Failed to try-submit-jobs")
        raise
    finally:
        cluster.demote_from_submitter()

    sys.exit(ret)



""""Factory functions for simulation configurations"""

import logging

from jade.exceptions import InvalidParameter
from jade.extensions.registry import Registry, ExtensionClassType
from jade.jobs.job_configuration import JobConfiguration
from jade.result import ResultsSummary
from jade.utils.utils import dump_data, load_data
from jade.utils.timing_utils import timed_debug


logger = logging.getLogger(__name__)


@timed_debug
def create_config_from_file(filename, **kwargs):
    """Create instance of a JobConfiguration from a config file.

    Returns
    -------
    JobConfiguration

    """
    data = load_data(filename)
    format = data.get("format_version", None)
    if format is None:
        upgrade_config_file(data, filename)

    return deserialize_config(data, **kwargs)


@timed_debug
def create_config_from_previous_run(config_file, output, result_type="successful", **kwargs):
    """Create instance of a JobConfiguration from a previous config file,
    returning only those of the type given

    Parameters
    ----------
    config_file : str
        location of config
    output : str
        location of previous results
    result_type : string
        type of results

    Returns
    -------
    JobConfiguration

    Raises
    ------
    InvalidParameter
            Raised if result_type is not successful or failed

    """
    allowed_types = ["successful", "failed", "missing"]
    if result_type not in allowed_types:
        raise InvalidParameter(f"given result type invalid: {result_type}")

    config = deserialize_config(load_data(config_file))
    summary = ResultsSummary(output)
    results_of_type = []

    if result_type == "successful":
        results_of_type = summary.get_successful_results()
    elif result_type == "failed":
        results_of_type = summary.get_failed_results()
    elif result_type == "missing":
        results_of_type = summary.get_missing_jobs(config.iter_jobs())

    parameters = []
    # Note that both jobs and results have `.name`.
    for result in results_of_type:
        job_parameters = config.get_job(result.name)
        parameters.append(job_parameters)

    config.reconfigure_jobs(parameters)
    return deserialize_config(config.serialize(), **kwargs)


@timed_debug
def deserialize_config(data, **kwargs):
    """Create instance of a JobConfiguration from a dict.

    Parameters
    ----------
    data : dict
        Dictionary loaded from a serialized config file.

    Returns
    -------
    JobConfiguration

    """
    registry = Registry()
    config_module = data["configuration_module"]
    config_class = data["configuration_class"]
    for ext in registry.iter_extensions():
        ext_cfg_class = ext[ExtensionClassType.CONFIGURATION]
        if ext_cfg_class.__module__ == config_module and ext_cfg_class.__name__ == config_class:
            return ext_cfg_class.deserialize(data, **kwargs)

    raise InvalidParameter(f"Cannot deserialize {config_module}.{config_class}")


def upgrade_config_file(data, filename):
    """Upgrades v0.1.0 format to the latest."""
    if data["class"] != "GenericCommandConfiguration":
        raise Exception(f"{filename} has an old format and must be regenerated")

    data["configuration_module"] = "jade.extensions.generic_command.generic_command_configuration"
    data["configuration_class"] = "GenericCommandConfiguration"
    data["format_version"] = JobConfiguration.FORMAT_VERSION
    data.pop("class")
    data.pop("extension")
    for job in data["jobs"]:
        job["extension"] = "generic_command"
        job["append_job_name"] = False
        job["append_output_dir"] = False
    dump_data(data, filename, indent=2)
    logger.info("Upgraded config file format: %s", filename)



"""CLI to run jobs."""

import itertools
import logging
from pathlib import Path

import click

from jade.loggers import setup_logging


logger = logging.getLogger(__name__)


@click.command()
@click.argument(
    "output",
    type=click.Path(exists=True),
)
def prune_files(output):
    """Deletes intermediate scripts, config files, and log files that are not needed if the job
    results were successful."""
    setup_logging(__name__, None, console_level=logging.INFO)
    base_path = Path(output)

    count = 0
    for path in itertools.chain(
        # Keep submit_jobs.log* files because there aren't many of them and they are useful.
        base_path.glob("submit_jobs_events*.log*"),
        base_path.glob("run_jobs_batch*.log*"),
        base_path.glob("config_batch*.json"),
        base_path.glob("*.sh"),
    ):
        path.unlink()
        count += 1
    print(f"Deleted {count} files from {output}.")



"""Defines a Result object."""

import logging
import os
from collections import namedtuple
from datetime import datetime
from time import time

from prettytable import PrettyTable

from jade.common import RESULTS_FILE
from jade.enums import JobCompletionStatus
from jade.exceptions import InvalidConfiguration, InvalidParameter, ExecutionError
from jade.utils.utils import load_data


logger = logging.getLogger(__name__)


class Result(
    namedtuple("Result", "name, return_code, status, exec_time_s, completion_time, hpc_job_id")
):
    """
    Result class containing data after jobs have finished. `completion_time`
    will be populated when result created and passed in when deserializing

    Attributes
    ----------
    name : str
    return_code : int
    status : str
    exec_time_s : int
    completion_time : int (default current timestamp)

    """

    def __new__(
        cls, name, return_code, status, exec_time_s, completion_time=None, hpc_job_id=None
    ):
        # add default values
        if completion_time is None:
            completion_time = time()
        if isinstance(status, JobCompletionStatus):
            status = status.value
        return super(Result, cls).__new__(
            cls, name, return_code, status, exec_time_s, completion_time, hpc_job_id
        )

    def is_canceled(self):
        """Return True if the result was canceled."""
        return self.return_code != 0 and self.status == JobCompletionStatus.CANCELED.value

    def is_failed(self):
        """Return True if the result was failed."""
        return self.return_code != 0 and self.status == JobCompletionStatus.FINISHED.value

    def is_successful(self):
        """Return True if the result was successful."""
        return self.return_code == 0 and self.status == JobCompletionStatus.FINISHED.value


def serialize_result(result):
    """Serialize a Result to a dict.

    Parameters
    ----------
    result : Result

    Returns
    -------
    dict

    """
    data = result._asdict()
    return data


def serialize_results(results):
    """Serialize a list of Result objects.

    Parameters
    ----------
    result : Result

    Returns
    -------
    list of dict

    """
    return [serialize_result(x) for x in results]


def deserialize_result(data):
    """Deserialize a Result from raw data.

    Parameters
    ----------
    data : dict

    Returns
    -------
    Result

    """
    hpc_job_id = data.get("hpc_job_id")
    if hpc_job_id == "None":
        hpc_job_id = None

    if "completion_time" in data.keys():
        return Result(
            data["name"],
            data["return_code"],
            data["status"],
            data["exec_time_s"],
            completion_time=data["completion_time"],
            hpc_job_id=hpc_job_id,
        )

    return Result(
        data["name"],
        data["return_code"],
        data["status"],
        data["exec_time_s"],
        hpc_job_id=hpc_job_id,
    )


def deserialize_results(data):
    """Deserialize a list of Result objects from raw data.

    Parameters
    ----------
    data : dict
        name to Result

    Returns
    -------
    dict of Result instances

    """
    return {x["name"]: deserialize_result(x) for x in data}


class ResultsSummary:
    """Provides summary of all job results."""

    def __init__(self, output_dir):
        self._output_dir = output_dir

        self._results_file = os.path.join(output_dir, RESULTS_FILE)
        if not os.path.exists(self._results_file):
            raise InvalidConfiguration(f"There is no results file in {output_dir}")

        data = self._parse(self._results_file)
        data["results"] = deserialize_results(data["results"])
        self._results = data
        self._missing_jobs = data["missing_jobs"]
        self._base_directory = data["base_directory"]

    def __iadd__(self, other):
        for result in other.results:
            self._results["results"][result.name] = result
        self._missing_jobs += other.missing_jobs
        return self

    @property
    def base_directory(self):
        """Return the base directory for the job results.

        Returns
        -------
        str

        """
        return self._base_directory

    @property
    def missing_jobs(self):
        return self._missing_jobs

    @property
    def results(self):
        """Return the results.

        Returns
        -------
        dict_values
            iterable of Result objects

        """
        return self._results["results"].values()

    @staticmethod
    def _parse(results_file):
        return load_data(results_file)

    def get_result(self, job_name):
        """Return the job result from the results
        Parameters
        ----------
        job_name : str
        Returns
        -------
        dict

        """
        return self._results["results"].get(job_name)

    def get_successful_result(self, job_name):
        """Return the successful job result from the results
        Parameters
        ----------
        job_name : str
        Returns
        -------
        dict
        Raises
        ------
        InvalidParameter
            Raised if job_name is not found.
        ExecutionError
            Raised if the result was not successful.
        """
        result = self.get_result(job_name)
        if result is None:
            raise InvalidParameter(f"result not found {job_name}")

        if not result.is_successful():
            raise ExecutionError(f"result wasn't successful: {result}")

        return result

    def get_results_by_type(self):
        """Return the results split by type.

        Returns
        -------
        dict

        """
        successful = []
        failed = []
        canceled = []

        for result in self._results["results"].values():
            if result.is_successful():
                successful.append(result)
            elif result.is_failed():
                failed.append(result)
            elif result.is_canceled():
                canceled.append(result)

        return {"successful": successful, "failed": failed, "canceled": canceled}

    def get_successful_results(self):
        """Return the successful results."""
        return [x for x in self._results["results"].values() if x.is_successful()]

    def get_missing_jobs(self, expected_jobs):
        """Return the jobs for which there are no results.

        Parameters
        ----------
        expected_jobs : list
        jobs that are expected to be in results

        Returns
        -------
        list

        """
        missing_jobs = []
        for job in expected_jobs:
            if self.get_result(job.name) is None:
                missing_jobs.append(job)

        return missing_jobs

    def get_canceled_results(self):
        """Return the canceled results."""
        return [x for x in self._results["results"].values() if x.is_canceled()]

    def get_failed_results(self):
        """Return the failed results."""
        return [x for x in self._results["results"].values() if x.is_failed()]

    def list_results(self):
        """Return the results.

        Returns
        -------
        list

        """
        return list(self._results["results"].values())

    def show_results(self, only_failed=False, only_successful=False):
        """Show the results in a table."""
        if only_successful and only_failed:
            raise InvalidParameter("only_failed and only_successful are mutually exclusive")

        print(f"Results from directory: {self._output_dir}")
        print(f"JADE Version: {self._results['jade_version']}")
        print(f"{self._results['timestamp']}\n")

        num_successful = 0
        num_failed = 0
        num_canceled = 0
        table = PrettyTable()
        table.field_names = [
            "Job Name",
            "Return Code",
            "Status",
            "Execution Time (s)",
            "Completion Time",
            "HPC Job ID",
        ]
        if self._results["results"]:
            first = next(iter(self._results["results"].values()))
            min_exec = first.exec_time_s
            max_exec = first.exec_time_s
        else:
            min_exec = 0.0
            max_exec = 0.0
        exec_times = []
        for result in self._results["results"].values():
            if result.is_successful():
                num_successful += 1
            elif result.is_failed():
                num_failed += 1
            else:
                assert result.is_canceled()
                num_canceled += 1
            if only_failed and result.return_code == 0:
                continue
            if only_successful and result.return_code != 0:
                continue
            if result.exec_time_s < min_exec:
                min_exec = result.exec_time_s
            if result.exec_time_s > max_exec:
                max_exec = result.exec_time_s
            exec_times.append(result.exec_time_s)
            table.add_row(
                [
                    result.name,
                    result.return_code,
                    result.status,
                    result.exec_time_s,
                    datetime.fromtimestamp(result.completion_time),
                    result.hpc_job_id,
                ]
            )

        num_missing = len(self._missing_jobs)
        total = num_successful + num_failed + num_canceled + num_missing
        assert total == len(self._results["results"]) + num_missing
        if exec_times:
            avg_exec = sum(exec_times) / len(exec_times)
        else:
            avg_exec = 0.0

        print(table)
        print(f"\nNum successful: {num_successful}")
        print(f"Num failed: {num_failed}")
        print(f"Num canceled: {num_canceled}")
        print(f"Num missing: {num_missing}")
        print(f"Missing job names: {self._missing_jobs}")
        print(f"Total: {total}\n")
        print("Avg execution time (s): {:.2f}".format(avg_exec))
        print("Min execution time (s): {:.2f}".format(min_exec))
        print("Max execution time (s): {:.2f}\n".format(max_exec))



"""Defines interface for async jobs."""

import abc


class AsyncJobInterface(abc.ABC):
    """Defines interface for async jobs."""

    @abc.abstractmethod
    def cancel(self):
        """Cancel the job."""

    @property
    @abc.abstractmethod
    def cancel_on_blocking_job_failure(self):
        """Cancel the job if a blocking job fails."""

    @abc.abstractmethod
    def get_id(self):
        """Return the ID of the job.

        Returns
        -------
        int

        """

    @abc.abstractmethod
    def is_complete(self):
        """Return True if the job is complete.

        Note: this must be called until the job actually completes.

        Returns
        -------
        bool

        Raises
        ------
        ExecutionError
            Raised if statuses cannot be retrieved.

        """

    @property
    @abc.abstractmethod
    def name(self):
        """Return the job name.

        Returns
        -------
        str

        """

    @property
    @abc.abstractmethod
    def return_code(self):
        """Returns the return code of the job.

        Returns
        -------
        int
            0 means success, otherwise failure

        """

    @abc.abstractmethod
    def run(self):
        """Run the job.

        Returns
        -------
        Status
            Returns GOOD if the job was successfully started.

        """

    @abc.abstractmethod
    def get_blocking_jobs(self):
        """Return the job names blocking this job.

        Returns
        -------
        set
            Empty set means that the job is not blocked.

        """

    @abc.abstractmethod
    def remove_blocking_job(self, name):
        """Remove the name from the job's blocking list.

        Parameters
        ----------
        name : str
            name of job that is now finished

        """

    @abc.abstractmethod
    def set_blocking_jobs(self, jobs):
        """Set the blocking jobs.

        Parameters
        ----------
        jobs : set
            names that are blocking

        """



"""Defines custom click options for cli commands"""

import click


class CustomOptions(click.Option):
    """Custom option class extending base click option"""

    def __init__(self, *args, **kwargs):
        if "allowed_values" in kwargs:
            self.allowed_values = kwargs.pop("allowed_values")

        if "not_required_if" in kwargs:
            self.not_required_if = kwargs.pop("not_required_if")

        if "required_if" in kwargs:
            self.required_if = kwargs.pop("required_if")

        super(CustomOptions, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        self.handle_custom_options(opts)

        return super(CustomOptions, self).handle_parse_result(ctx, opts, args)

    def handle_custom_options(self, opts):
        """Handles custom options that have been created"""
        if hasattr(self, "allowed_values") and isinstance(self.allowed_values, list):
            if self.name in opts and opts[self.name] not in self.allowed_values:
                raise ValueError(
                    "Invalid value given, only allowed values are " + f"{self.allowed_values}"
                )

        if hasattr(self, "not_required_if"):
            not_required_if_present = self.not_required_if in opts.keys()

            if not_required_if_present:
                self.required = False
                self.prompt = None

        if hasattr(self, "required_if"):
            required_if_present = self.required_if in opts.keys()

            if required_if_present and opts[self.required_if] is not None:
                self.required = True



"""CLI to collect and process resource stats."""

import logging
import os
import shutil
import signal
import sys
import time

import click
import psutil

from jade.events import EventsSummary
from jade.loggers import setup_event_logging
from jade.models.submitter_params import ResourceMonitorStats
from jade.resource_monitor import ResourceMonitorLogger
from jade.utils.run_command import check_run_command


logger = logging.getLogger(__name__)


@click.command()
@click.option(
    "-d",
    "--duration",
    default=None,
    type=int,
    show_default=True,
    help="Total time to collect resource stats. Default is infinite.",
)
@click.option(
    "-i",
    "--interval",
    default=1,
    type=int,
    show_default=True,
    help="Interval in seconds on which to collect resource stats.",
)
@click.option("-o", "--output", default="stats", show_default=True, help="Output directory.")
@click.option(
    "--plots/--no-plots",
    default=False,
    is_flag=True,
    show_default=True,
    help="Generate plots when collection is complete.",
)
@click.option(
    "--cpu/--no-cpu", default=True, is_flag=True, show_default=True, help="Enable CPU monitoring"
)
@click.option(
    "--disk/--no-disk",
    default=True,
    is_flag=True,
    show_default=True,
    help="Enable disk monitoring",
)
@click.option(
    "--memory/--no-memory",
    default=True,
    is_flag=True,
    show_default=True,
    help="Enable memory monitoring",
)
@click.option(
    "--network/--no-network",
    default=True,
    is_flag=True,
    show_default=True,
    help="Enable network monitoring",
)
@click.option(
    "-p",
    "--process-ids",
    multiple=True,
    show_default=True,
    type=int,
    help="Process IDs to monitor. Ex: -p 3114 -p 3115",
)
@click.option(
    "-u",
    "--process-ids-from-user",
    type=str,
    help="Monitor all process IDs from this username. Can be combined with --process-ids-from-substring",
)
@click.option(
    "-P",
    "--process-ids-from-substrings",
    type=str,
    multiple=True,
    help="Monitor all process IDs with command line exec + args that include one or more substrings.",
)
@click.option(
    "-r",
    "--refresh-process-IDs",
    is_flag=True,
    default=False,
    show_default=True,
    help="Refresh the process IDs every interval. Default behavior is to only monitor the "
    "original IDs.",
)
@click.option(
    "--children/--no-children",
    default=False,
    is_flag=True,
    show_default=True,
    help="Aggregate child process utilization.",
)
@click.option(
    "--recurse-children/--no-recurse-children",
    default=False,
    is_flag=True,
    show_default=True,
    help="Search for all child processes recursively.",
)
@click.option(
    "-f",
    "--force",
    default=False,
    is_flag=True,
    show_default=True,
    help="Delete output directory if it exists.",
)
@click.pass_context
def collect(
    ctx,
    duration,
    interval,
    output,
    plots,
    cpu,
    disk,
    memory,
    network,
    process_ids,
    process_ids_from_user,
    process_ids_from_substrings,
    refresh_process_ids,
    children,
    recurse_children,
    force,
):
    """Collect resource utilization stats. Stop collection by setting duration, pressing Ctrl-c,
    or sending SIGTERM to the process ID.
    """
    if os.path.exists(output):
        if force:
            shutil.rmtree(output)
        else:
            print(
                f"The directory {output} already exists. Delete it or run with --force",
                file=sys.stderr,
            )
            sys.exit(1)

    process = bool(process_ids or process_ids_from_user or process_ids_from_substrings)
    if process_ids and (process_ids_from_user or process_ids_from_substrings):
        print("Explicit process IDs and process ID filters cannot both be passed", file=sys.stderr)
        sys.exit(1)

    os.makedirs(output)
    event_file = os.path.join(output, "stats_events.log")
    setup_event_logging(event_file, mode="a")
    stats = ResourceMonitorStats(
        cpu=cpu,
        disk=disk,
        memory=memory,
        network=network,
        process=process,
        include_child_processes=children,
        recurse_child_processes=recurse_children,
    )

    if process_ids:
        process_id_mapping = get_process_id_mapping(process_ids)
    else:
        process_id_mapping = None

    monitor = ResourceMonitorLogger("ResourceMonitor", stats)
    start_time = time.time()

    show_cmd = f"jade stats show -o {output} [STATS]"
    print(f"Collecting stats. When complete run '{show_cmd}' to view stats.", file=sys.stderr)
    signal.signal(signal.SIGTERM, sigterm_handler)
    with ctx:
        try:
            while g_collect_stats:
                if stats.process:
                    if (process_ids_from_user or process_ids_from_substrings) and (
                        process_id_mapping is None or refresh_process_ids
                    ):
                        process_id_mapping = find_process_ids(
                            process_ids_from_user, process_ids_from_substrings
                        )
                monitor.log_resource_stats(ids=process_id_mapping)
                time.sleep(interval)
                if duration is not None and time.time() - start_time > duration:
                    print(f"Exceeded {duration} seconds. Exiting.", file=sys.stderr)
                    break
        except KeyboardInterrupt:
            print(" Detected Ctrl-c, exiting", file=sys.stderr)
        if plots:
            check_run_command(f"jade stats plot -o {output}")
        else:
            # This generates parquet files for each stat.
            EventsSummary(output)


def get_process_mapping_name(process: psutil.Process):
    """Return a mapping of name to pid. The name should be suitable for plots."""
    # Including the entire command line is often way too long.
    # name() is often better than the first arg (python instead of full path to python)
    # This tries to get the best of all worlds and ensure uniqueness.
    cmdline = process.cmdline()
    if len(cmdline) > 1:
        name = process.name() + " " + " ".join(cmdline[1:])
    else:
        name = process.name()

    if len(name) > 20:
        name = name[:20] + "..."

    return name + f" ({process.pid})"


def get_process_id_mapping(process_ids):
    process_id_mapping = {}
    for pid in process_ids:
        process = psutil.Process(pid)
        process_id_mapping[get_process_mapping_name(process)] = pid

    return process_id_mapping


def find_process_ids(user, substrings):
    process_id_mapping = {}
    for process in psutil.process_iter():
        if process.pid == os.getpid():
            # Don't track this process.
            continue
        try:
            if user is not None and process.username() != user:
                continue
            found_substring = False
            for substring in substrings:
                if substring in " ".join(process.cmdline()):
                    found_substring = True
            if not found_substring:
                continue
            process_id_mapping[get_process_mapping_name(process)] = process.pid
        except (psutil.NoSuchProcess, psutil.AccessDenied):
            pass

    return process_id_mapping


g_collect_stats = True


def sigterm_handler(signum, frame):
    global g_collect_stats
    print("Detected SIGTERM", file=sys.stderr)
    g_collect_stats = False



"""
CLI to show events of a scenario.
"""

import datetime
import json
import os
import sys
from collections import defaultdict
from pathlib import Path

import click
import numpy as np
from prettytable import PrettyTable
from psutil._common import bytes2human

from jade.cli.collect_stats import collect
from jade.common import EVENTS_DIR, OUTPUT_DIR, STATS_DIR
from jade.loggers import setup_logging
from jade.events import EventsSummary
from jade.resource_monitor import (
    CpuStatsViewer,
    DiskStatsViewer,
    MemoryStatsViewer,
    NetworkStatsViewer,
    ProcessStatsViewer,
)
from jade.utils.utils import dump_data, load_data


STATS = ("cpu", "disk", "mem", "net", "proc")


@click.group()
def stats():
    """Collect new stats or view stats from an existing run."""
    setup_logging("stats", None)


@click.argument("stats", nargs=-1)
@click.option(
    "-o",
    "--output",
    default=OUTPUT_DIR,
    show_default=True,
    help="JADE submission output directory.",
)
@click.command()
def plot(stats, output):
    """Plot stats from a run to files.

    \b
    Examples:
    jade stats plot
    jade stats plot cpu
    jade stats plot disk
    jade stats plot mem
    jade stats plot net
    jade stats plot cpu disk mem proc
    """
    events = EventsSummary(output)

    if not stats:
        stats = STATS

    plot_dir = Path(output) / STATS_DIR
    plot_dir.mkdir(exist_ok=True)
    for stat in stats:
        if stat == "cpu":
            viewer = CpuStatsViewer(events)
        elif stat == "disk":
            viewer = DiskStatsViewer(events)
        elif stat == "mem":
            viewer = MemoryStatsViewer(events)
        elif stat == "net":
            viewer = NetworkStatsViewer(events)
        elif stat == "proc":
            viewer = ProcessStatsViewer(events)
        else:
            print(f"Invalid stat={stat}", file=sys.stderr)
            sys.exit(1)
        viewer.plot_to_file(plot_dir)


@click.argument("stats", nargs=-1)
@click.option(
    "-j",
    "--json-summary",
    default=False,
    is_flag=True,
    show_default=True,
    help="Only show the summary stats per node in JSON format.",
)
@click.option("-o", "--output", default=OUTPUT_DIR, show_default=True, help="Output directory.")
@click.option(
    "-s",
    "--summary-only",
    default=False,
    is_flag=True,
    show_default=True,
    help="Only show the summary stats per node.",
)
@click.command()
def show(stats, json_summary, output, summary_only):
    """Shows stats from a run.

    \b
    Examples:
    jade stats show
    jade stats show cpu
    jade stats show disk
    jade stats show mem
    jade stats show net
    jade stats show cpu disk mem
    jade stats show --summary-only cpu mem proc
    jade stats show --json-summary cpu disk mem
    """
    events_path = Path(output) / EVENTS_DIR
    stats_path = Path(output) / STATS_DIR
    if not events_path.exists():
        print(f"{output} does not contain JADE stats", file=sys.stderr)
        sys.exit(1)
    if not stats:
        stats = STATS
    if stats_path.exists():
        json_files = list(stats_path.glob("*.json"))
        json_files.sort()
    else:
        json_files = []
    if json_files:
        _show_summary_stats(stats, json_summary, json_files)
    else:
        _show_periodic_stats(stats, json_summary, output, summary_only)
    return 0


def _show_summary_stats(stats, json_summary, json_files):
    type_mapping = {
        CpuStatsViewer.metric(): "cpu",
        DiskStatsViewer.metric(): "disk",
        MemoryStatsViewer.metric(): "mem",
        NetworkStatsViewer.metric(): "net",
        ProcessStatsViewer.metric(): "proc",
    }
    reverse_mapping = {v: k for k, v in type_mapping.items()}
    cls_mapping = {
        CpuStatsViewer.metric(): CpuStatsViewer,
        DiskStatsViewer.metric(): DiskStatsViewer,
        MemoryStatsViewer.metric(): MemoryStatsViewer,
        NetworkStatsViewer.metric(): NetworkStatsViewer,
        ProcessStatsViewer.metric(): ProcessStatsViewer,
    }
    filtered = []
    for filename in json_files:
        data = load_data(filename)
        for entry in data:
            if type_mapping[entry["type"]] in stats:
                filtered.append(entry)

    if json_summary:
        print(json.dumps(filtered, indent=2))
    else:
        by_type_and_batch = defaultdict(dict)
        for entry in filtered:
            if entry["batch"] not in by_type_and_batch[entry["type"]]:
                by_type_and_batch[entry["type"]][entry["batch"]] = []
            by_type_and_batch[entry["type"]][entry["batch"]].append(entry)
        for stat in stats:
            text = f"{reverse_mapping[stat]} statistics for each batch"
            print(f"\n{text}")
            print("=" * len(text) + "\n")

            stat_type = reverse_mapping[stat]
            stat_cls = cls_mapping[stat_type]
            for resource_type in by_type_and_batch:
                if resource_type != reverse_mapping[stat]:
                    continue
                for batch, entries in by_type_and_batch[resource_type].items():
                    for entry in entries:
                        if stat_type == "Process":
                            tag = entry["name"]
                        else:
                            tag = batch
                        table = PrettyTable(title=f"{stat_type} {tag} summary")
                        table.field_names = [stat] + list(entry["average"].keys())
                        row = ["Average"]
                        for field, val in entry["average"].items():
                            row.append(stat_cls.get_printable_value(field, val))
                        table.add_row(row)
                        row = ["Minimum"]
                        for field, val in entry["minimum"].items():
                            row.append(stat_cls.get_printable_value(field, val))
                        table.add_row(row)
                        row = ["Maximum"]
                        for field, val in entry["maximum"].items():
                            row.append(stat_cls.get_printable_value(field, val))
                        table.add_row(row)
                        print(table)
                        print()


def _show_periodic_stats(stats, json_summary, output, summary_only):
    events = EventsSummary(output)

    summaries_as_dicts = []
    for stat in stats:
        if stat == "cpu":
            viewer = CpuStatsViewer(events)
        elif stat == "disk":
            viewer = DiskStatsViewer(events)
        elif stat == "mem":
            viewer = MemoryStatsViewer(events)
        elif stat == "net":
            viewer = NetworkStatsViewer(events)
        elif stat == "proc":
            viewer = ProcessStatsViewer(events)
        else:
            print(f"Invalid stat={stat}", file=sys.stderr)
            sys.exit(1)
        if json_summary:
            summaries_as_dicts += viewer.get_stats_summary()
        else:
            viewer.show_stats(show_all_timestamps=not summary_only)

    if json_summary:
        print(json.dumps(summaries_as_dicts, indent=2, cls=NumpyJSONEncoder))


class NumpyJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.int64):
            return int(obj)
        if isinstance(obj, np.float64):
            return float(obj)
        return obj


@click.option(
    "--human-readable/--no-human-readable",
    is_flag=True,
    default=True,
    show_default=True,
    help="Output directory.",
)
@click.option("-o", "--output", default=OUTPUT_DIR, show_default=True, help="Output directory.")
@click.command()
def bytes_consumed(output, human_readable):
    events = EventsSummary(output)
    consumed = events.get_bytes_consumed()
    if human_readable:
        print(bytes2human(consumed))
    else:
        print(consumed)


@click.option(
    "--human-readable/--no-human-readable",
    is_flag=True,
    default=True,
    show_default=True,
    help="Output directory.",
)
@click.option("-o", "--output", default=OUTPUT_DIR, show_default=True, help="Output directory.")
@click.command()
def exec_time(output, human_readable):
    events = EventsSummary(output)
    config_exec_time = events.get_config_exec_time()
    if human_readable:
        print(datetime.timedelta(seconds=config_exec_time))
    else:
        print(config_exec_time)


stats.add_command(bytes_consumed)
stats.add_command(collect)
stats.add_command(exec_time)
stats.add_command(plot)
stats.add_command(show)



"""Utility functions for timing measurements."""

import functools
import logging
import time


logger = logging.getLogger(__name__)


def timed_info(func):
    """Decorator to measure and logger.info a function's execution time."""

    @functools.wraps(func)
    def timed_(*args, **kwargs):
        return _timed(func, logger.info, *args, **kwargs)

    return timed_


def timed_debug(func):
    """Decorator to measure and logger.debug a function's execution time."""

    @functools.wraps(func)
    def timed_(*args, **kwargs):
        return _timed(func, logger.debug, *args, **kwargs)

    return timed_


def _timed(func, log_func, *args, **kwargs):
    start = time.time()
    result = func(*args, **kwargs)
    total = time.time() - start
    log_func("execution-time=%s func=%s", get_time_duration_string(total), func.__name__)
    return result


def get_time_duration_string(seconds):
    """Returns a string with the time converted to reasonable units."""
    if seconds >= 1:
        val = "{:.3f} s".format(seconds)
    elif seconds >= 0.001:
        val = "{:.3f} ms".format(seconds * 1000)
    elif seconds >= 0.000001:
        val = "{:.3f} us".format(seconds * 1000000)
    elif seconds == 0:
        val = "0 s"
    else:
        val = "{:.3f} ns".format(seconds * 1000000000)

    return val


class TimerStats:
    """Tracks timing stats for one code block."""

    def __init__(self, name):
        self._name = name
        self._count = 0
        self._max = 0.0
        self._min = None
        self._avg = 0.0
        self._total = 0.0

    def get_stats(self):
        """Get the current stats summary.

        Returns
        -------
        dict

        """
        avg = 0 if self._count == 0 else self._total / self._count
        return {
            "min": self._min,
            "max": self._max,
            "total": self._total,
            "avg": avg,
            "count": self._count,
        }

    def log_stats(self):
        """Log a summary of the stats."""
        if self._count == 0:
            logger.info("No stats have been recorded for %s.", self._name)
            return

        x = self.get_stats()
        text = "total={:.3f}s avg={:.3f}ms max={:.3f}ms min={:.3f}ms count={}".format(
            x["total"], x["avg"] * 1000, x["max"] * 1000, x["min"] * 1000, x["count"]
        )
        logger.info("TimerStats summary: %s: %s", self._name, text)

    def update(self, duration):
        """Update the stats with a new timing."""
        self._count += 1
        self._total += duration
        if duration > self._max:
            self._max = duration
        if self._min is None or duration < self._min:
            self._min = duration


class Timer:
    """Times a code block."""

    def __init__(self, timer_stats, name):
        self._start = None
        self._timer_stat = timer_stats.get_stat(name)

    def __enter__(self):
        self._start = time.perf_counter()

    def __exit__(self, exc, value, tb):
        self._timer_stat.update(time.perf_counter() - self._start)


def track_timing(collector):
    """Decorator to track statistics on a function's execution time.

    Parameters
    ----------
    collector : TimerStatsCollector

    """

    def wrap(func):
        def timed_(*args, **kwargs):
            return _timed_func(collector, func, *args, **kwargs)

        return timed_

    return wrap


def _timed_func(timer_stats, func, *args, **kwargs):
    with Timer(timer_stats, func.__qualname__):
        return func(*args, **kwargs)


class TimerStatsCollector:
    """Collects statistics for timed code segments."""

    def __init__(self):
        self._stats = {}

    def clear(self):
        """Clear all stats."""
        self._stats.clear()

    def get_stat(self, name):
        """Return a TimerStats.

        Parameters
        ----------
        name : str

        Returns
        -------
        TimerStats

        """
        if name not in self._stats:
            self.register_stat(name)
        return self._stats[name]

    def log_stats(self, clear=False):
        """Log statistics for all tracked stats.

        Parameters
        ----------
        clear : bool
            If True, clear all stats.

        """
        for stat in self._stats.values():
            stat.log_stats()
        if clear:
            self._stats.clear()

    def register_stat(self, name):
        """Register tracking of a new stat.

        Parameters
        ----------
        name : str

        Returns
        -------
        TimerStats

        """
        assert name not in self._stats
        stat = TimerStats(name)
        self._stats[name] = stat
        return stat


timer_stats_collector = TimerStatsCollector()



"""Performs postprocessing and analysis on job results."""

import logging
import os

from jade.common import JOBS_OUTPUT_DIR
from jade.enums import JobCompletionStatus
from jade.exceptions import ExecutionError, InvalidParameter
from jade.result import ResultsSummary
from jade.utils.utils import load_data


logger = logging.getLogger(__name__)


class JobAnalysis:
    """Provides functionality to analyze job results."""

    def __init__(self, output_dir, config):
        self._output_dir = output_dir
        self._config = config
        self._results = ResultsSummary(output_dir)

    def get_job(self, job_name):
        """Return the job from the config file with job_name.

        Parameters
        ----------
        job_name : str

        Returns
        -------
        namedtuple

        """
        return self._config.get_job(job_name)

    def get_successful_result(self, job_name):
        """Return the job result from the results file.
        Refer to :func:`~jade.result.ResultSummary.get_successful_result`.

        """
        return self._results.get_successful_result(job_name)

    def get_simulation(self, job_name):
        """Return a simulation object for the job_name.

        Parameters
        ----------
        job_name : str

        Returns
        -------
        JobExecutionInterface

        """
        # Make sure it was successful, otherwise it will raise.
        self.get_successful_result(job_name)

        job = self.get_job(job_name)
        simulation = self._config.create_from_result(
            job, os.path.join(self._output_dir, JOBS_OUTPUT_DIR)
        )
        return simulation

    def list_results(self):
        """Return a list of Result objects."""
        return self._results.list_results()

    @property
    def output_dir(self):
        return self._output_dir

    def show_results(self, only_failed=False, only_successful=False):
        """Show the results in terminal."""
        return self._results.show_results(only_failed=only_failed, only_successful=only_successful)


def get_result(results_file, job_name):
    """Return the job result from the results file.

    Parameters
    ----------
    results_file : str
    job_name : str

    Returns
    -------
    dict

    Raises
    ------
    InvalidParameter
        Raised if job_name is not found.

    """
    results = load_data(results_file)
    for result in results:
        if result.name == job_name:
            return result

    raise InvalidParameter(f"result not found {job_name}")


def get_successful_result(results_file, job_name):
    """Return the job result from the results file.

    Parameters
    ----------
    results_file : str
    job_name : str

    Returns
    -------
    dict

    Raises
    ------
    InvalidParameter
        Raised if job_name is not found.
    ExecutionError
        Raised if the result was not successful.

    """
    result = get_result(results_file, job_name)
    if not result.is_successful():
        raise ExecutionError(f"result was not successful: {result}")

    return result



"""
Script for creating inputs.json file
"""
import json


COUNTRIES = [
    "Australia",
    "Brazil",
    "Canada",
    "China",
    "France",
    "Germany",
    "India",
    "Italy",
    "Japan",
    "Russia",
    "United Kingdom",
    "United States",
]

INPUTS_FILE = "inputs.json"


def main():
    """Create inputs.json file"""
    inputs = []
    for country in COUNTRIES:
        filename = country.replace(" ", "_").lower() + ".csv"
        parameter = {"country": country, "data": f"gdp/countries/{filename}"}

        # Generate wrong path for raising error while running job
        if country in ["Australia", "Canada"]:
            parameter["data"] = "x" + parameter["data"]

        inputs.append(parameter)

    with open(INPUTS_FILE, "w") as f:
        json.dump(inputs, f, indent=2)

    print("inputs.json created.")


if __name__ == "__main__":
    main()



"""CLI to show job status."""

import logging
import os
import socket
import sys

import click
from prettytable import PrettyTable

from jade.common import OUTPUT_DIR
from jade.exceptions import InvalidConfiguration
from jade.jobs.cluster import Cluster
from jade.hpc.common import HpcJobStatus
from jade.hpc.hpc_manager import HpcManager
from jade.loggers import setup_logging
from jade.models.submission_group import make_submission_group_lookup
from jade.utils.subprocess_manager import run_command


logger = logging.getLogger(__name__)


@click.command()
@click.option(
    "-o",
    "--output",
    default=OUTPUT_DIR,
    show_default=True,
    type=click.Path(exists=True),
    help="directory containing submission output",
)
@click.option(
    "-j",
    "--job-status",
    is_flag=True,
    default=False,
    show_default=True,
    help="include individual job status",
)
@click.option(
    "-n",
    "--no-prompts",
    is_flag=True,
    default=False,
    show_default=True,
    help="Do not prompt to confirm to run try-submit-jobs if the submission is incomplete.",
)
@click.option(
    "--verbose", is_flag=True, default=False, show_default=True, help="Enable verbose log output."
)
def show_status(output, job_status, no_prompts, verbose):
    """Shows the status of active HPC jobs."""
    level = logging.DEBUG if verbose else logging.INFO
    setup_logging(__name__, None, console_level=level)
    try:
        cluster, _ = Cluster.deserialize(output, deserialize_jobs=True)
    except InvalidConfiguration:
        print(f"{output} is not a JADE output directory used in cluster mode", file=sys.stderr)
        sys.exit(1)

    summary = cluster.get_status_summary(include_jobs=job_status)
    print(f"Summary of jobs running in path={output}:")
    for key, val in summary.items():
        if key == "job_status":
            print(f"  active_hpc_job_ids: {val['hpc_job_ids']}")
        else:
            print(f"  {key}: {val}")

    if job_status and summary["job_status"]["jobs"]:
        print("\nJob Status:")
        jobs = summary["job_status"]["jobs"]
        table = PrettyTable()
        table.field_names = list(jobs[0].keys())
        for job in jobs:
            row = []
            for name in table.field_names:
                if name == "blocked_by":
                    row.append(" ".join(job[name]))
                elif name == "state":
                    row.append(job[name].value)
                else:
                    row.append(job[name])
            table.add_row(row)
        print(table)

    if not cluster.is_complete():
        # Check if the last job got killed or timed-out and try to restart.
        run_new_submitter = False
        if cluster.job_status.hpc_job_ids:
            groups = make_submission_group_lookup([cluster.config.submission_groups[0]])
            hpc_mgr = HpcManager(groups, output)
            all_jobs_are_none = True
            for job_id in cluster.job_status.hpc_job_ids:
                status = hpc_mgr.check_status(job_id=job_id)
                if status != HpcJobStatus.NONE:
                    all_jobs_are_none = False
                    break
            if all_jobs_are_none:
                logger.warn("HPC job statuses may be out-of-date. Jobs may have timed-out.")
                run_new_submitter = True
        else:
            logger.error(
                "Jobs are not complete but there no active HPC jobs. Jobs may have timed-out."
            )
            run_new_submitter = True
        if run_new_submitter:
            if not no_prompts:
                while True:
                    resp = input(
                        f"""\n'jade try-submit-jobs' must be run. If all jobs are complete then
this will generate reports, which may consume lots of compute resources, 
especially if you are collecting periodic resource utilization stats. 
You may not want to run this command on the login node. You are currently
logged into hostname={socket.gethostname()}.

Continue? [N] >>> """
                    ).lower()
                    if resp in ("n", ""):
                        print("Exiting", file=sys.stderr)
                        sys.exit(0)
                    elif resp == "y":
                        break
                    else:
                        print(f"\nERROR: Did not understand input={resp}")

            try_submit_cmd = f"jade try-submit-jobs {output}"
            if verbose:
                try_submit_cmd += " --verbose"
            run_command(try_submit_cmd)



"""CLI to show results of a scenario."""

import logging
import sys
import click
from pathlib import Path

from jade.common import OUTPUT_DIR
from jade.exceptions import InvalidConfiguration
from jade.jobs.job_post_process import JobPostProcess
from jade.loggers import setup_logging
from jade.result import ResultsSummary


@click.option(
    "-f",
    "--failed",
    is_flag=True,
    default=False,
    show_default=True,
    help="Show failed results only.",
)
@click.option("-o", "--output", default=OUTPUT_DIR, show_default=True, help="Output directory.")
@click.option(
    "-s",
    "--successful",
    is_flag=True,
    default=False,
    show_default=True,
    help="Show successful results only.",
)
@click.option(
    "-p",
    "--post-process",
    is_flag=True,
    default=False,
    show_default=True,
    help="Show post-process results.",
)
@click.option("-j", "--job-name", default=None, help="Specific job to show post-process results.")
@click.option(
    "--verbose", is_flag=True, default=False, show_default=True, help="Enable verbose log output."
)
@click.command()
def show_results(failed, output, successful, post_process, job_name, verbose):
    """Shows the results of a batch of jobs."""
    if not Path(output).exists():
        print(f"{output} does not exist", file=sys.stderr)
        sys.exit(1)

    level = logging.DEBUG if verbose else logging.WARNING
    setup_logging("show_results", None, console_level=level)

    if post_process:
        JobPostProcess.show_results(output, job_name)
        sys.exit(0)

    try:
        results = ResultsSummary(output)
    except InvalidConfiguration:
        print(
            f"No results are available in {output}. To check status of in-progress jobs run "
            f"'jade show-status -o {output}'",
            file=sys.stderr,
        )
        sys.exit(1)

    results.show_results(only_failed=failed, only_successful=successful)



"""Implements a container for jobs represented by a key."""

import logging
import random

from jade.exceptions import InvalidConfiguration, InvalidParameter
from jade.jobs.job_container_interface import JobContainerInterface


logger = logging.getLogger(__name__)


class JobContainerByName(JobContainerInterface):
    """Stores jobs by name, which must be unique."""

    def __init__(self):
        # name: JobParametersInterface
        self._jobs = {}

    def __iter__(self):
        for job in self._jobs.values():
            yield job

    def __len__(self):
        return len(self._jobs)

    def add_job(self, job):
        if job.name in self._jobs:
            raise InvalidConfiguration(f"job name {job.name} is already stored")
        self._jobs[job.name] = job
        logger.debug("Added job %s", job.name)

    def clear(self):
        self._jobs.clear()
        logger.debug("Cleared all jobs.")

    def remove_job(self, job):
        self._jobs.pop(job.name)
        logger.info("Removed job %s", job.name)

    def get_job(self, name):
        job = self._jobs.get(name)
        if job is None:
            raise InvalidParameter(f"job={name} is not stored")
        return self._jobs[name]

    def get_jobs(self, sort=False):
        if sort:
            names = list(self._jobs.keys())
            names.sort()
            return [self._jobs[x] for x in names]

        return list(self)

    def shuffle(self):
        names = list(self._jobs.keys())
        random.shuffle(names)
        new_jobs = {}
        for name in names:
            new_jobs[name] = self._jobs.pop(name)
        self._jobs = new_jobs



"""Cluster-related CLI commands"""

import logging
import sys

import click

from jade.common import OUTPUT_DIR
from jade.exceptions import InvalidConfiguration
from jade.jobs.cluster import Cluster
from jade.hpc.hpc_manager import HpcManager
from jade.loggers import setup_logging
from jade.models.submission_group import make_submission_group_lookup


logger = logging.getLogger(__name__)


@click.command()
@click.argument("output_dir", type=click.Path(exists=True))
@click.option(
    "--verbose", is_flag=True, default=False, show_default=True, help="Enable verbose log output."
)
def am_i_manager(output_dir, verbose):
    """Print 'true' or 'false' depending on whether the current node is the manager node."""
    level = logging.DEBUG if verbose else logging.INFO
    setup_logging(__name__, None, console_level=level)
    try:
        cluster, _ = Cluster.deserialize(output_dir, deserialize_jobs=True)
    except InvalidConfiguration:
        print(f"{output_dir} is not a JADE output directory used in cluster mode", file=sys.stderr)
        sys.exit(1)

    if cluster.is_complete():
        print("All jobs are already complete.")
        sys.exit()

    groups = make_submission_group_lookup([cluster.config.submission_groups[0]])
    hpc_mgr = HpcManager(groups, output_dir)
    am_manager = hpc_mgr.am_i_manager()
    print(str(am_manager).lower(), end="")


@click.command()
@click.argument("output_dir", type=click.Path(exists=True))
@click.option(
    "-j",
    "--job-id",
    type=str,
    help="If set, only get hostnames for this HPC job ID. Default is all HPC jobs.",
)
@click.option(
    "--verbose", is_flag=True, default=False, show_default=True, help="Enable verbose log output."
)
def hostnames(output_dir, job_id, verbose):
    """Show the hostnames of active nodes participating in the batch."""
    level = logging.DEBUG if verbose else logging.INFO
    setup_logging(__name__, None, console_level=level)
    try:
        cluster, _ = Cluster.deserialize(output_dir, deserialize_jobs=True)
    except InvalidConfiguration:
        print(f"{output_dir} is not a JADE output directory used in cluster mode", file=sys.stderr)
        sys.exit(1)

    if cluster.is_complete():
        print("All jobs are already complete.")
        sys.exit()

    groups = make_submission_group_lookup([cluster.config.submission_groups[0]])
    hpc_mgr = HpcManager(groups, output_dir)
    nodes = []
    for _job_id in cluster.job_status.hpc_job_ids:
        if job_id is not None and _job_id != job_id:
            continue
        nodes += hpc_mgr.list_active_nodes(_job_id)

    if not nodes:
        print("No nodes were detected.", file=sys.stderr)
        sys.exit(1)

    print(" ".join(nodes))


@click.command()
@click.argument("output_dir", type=click.Path(exists=True))
@click.argument("job_id", type=str)
@click.option(
    "--verbose", is_flag=True, default=False, show_default=True, help="Enable verbose log output."
)
def manager_node(output_dir, job_id, verbose):
    """Print the name of the manager node to the console. Requires a single job in the batch."""
    level = logging.DEBUG if verbose else logging.INFO
    setup_logging(__name__, None, console_level=level)
    try:
        cluster, _ = Cluster.deserialize(output_dir, deserialize_jobs=True)
    except InvalidConfiguration:
        print(f"{output_dir} is not a JADE output directory used in cluster mode", file=sys.stderr)
        sys.exit(1)

    if cluster.is_complete():
        print("All jobs are already complete.")
        sys.exit()

    groups = make_submission_group_lookup([cluster.config.submission_groups[0]])
    hpc_mgr = HpcManager(groups, output_dir)
    if job_id not in cluster.job_status.hpc_job_ids:
        print(f"job_id={job_id} is not active", file=sys.stderr)
        sys.exit(1)

    node = hpc_mgr.list_active_nodes(job_id)[0]
    print(node)


@click.group()
def cluster():
    pass


cluster.add_command(am_i_manager)
cluster.add_command(hostnames)
cluster.add_command(manager_node)



"""SLURM management functionality"""

import logging
import os
import re
from datetime import datetime, timedelta

from jade.enums import Status
from jade.exceptions import ExecutionError  # , InvalidConfiguration
from jade.hpc.common import HpcJobStats, HpcJobStatus, HpcJobInfo
from jade.hpc.hpc_manager_interface import HpcManagerInterface
from jade.utils.run_command import check_run_command, run_command
from jade.utils import utils


logger = logging.getLogger(__name__)


class SlurmManager(HpcManagerInterface):
    """Manages Slurm jobs."""

    _STATUSES = {
        "PENDING": HpcJobStatus.QUEUED,
        "CONFIGURING": HpcJobStatus.QUEUED,
        "RUNNING": HpcJobStatus.RUNNING,
        "COMPLETED": HpcJobStatus.COMPLETE,
        "COMPLETING": HpcJobStatus.COMPLETE,
    }
    _REGEX_SBATCH_OUTPUT = re.compile(r"Submitted batch job (\d+)")

    def __init__(self, config):
        self._config = config

    def am_i_manager(self):
        return os.environ.get("SLURM_NODEID", 1) == "0"

    def cancel_job(self, job_id):
        return run_command(f"scancel {job_id}")

    def check_status(self, name=None, job_id=None):
        field_names = ("jobid", "name", "state")
        cmd = f"squeue -u {self.USER} --Format \"{','.join(field_names)}\" -h"
        if name is not None:
            cmd += f" -n {name}"
        elif job_id is not None:
            cmd += f" -j {job_id}"
        else:
            # Mutual exclusivity should be handled in HpcManager.
            assert False

        output = {}
        # Transient failures could be costly. Retry for up to one minute.
        errors = ["Invalid job id specified"]
        ret = run_command(cmd, output, num_retries=6, retry_delay_s=10, error_strings=errors)
        if ret != 0:
            if "Invalid job id specified" in output["stderr"]:
                return HpcJobInfo("", "", HpcJobStatus.NONE)

            logger.error(
                "Failed to run squeue command=[%s] ret=%s err=%s", cmd, ret, output["stderr"]
            )
            raise ExecutionError(f"squeue command failed: {ret}")

        stdout = output["stdout"]
        logger.debug("squeue output:  [%s]", stdout)
        fields = stdout.split()
        if not fields:
            # No jobs are currently running.
            return HpcJobInfo("", "", HpcJobStatus.NONE)

        assert len(fields) == len(field_names)
        job_info = HpcJobInfo(
            fields[0], fields[1], self._STATUSES.get(fields[2], HpcJobStatus.UNKNOWN)
        )
        return job_info

    def check_statuses(self):
        field_names = ("jobid", "state")
        cmd = f"squeue -u {self.USER} --Format \"{','.join(field_names)}\" -h"

        output = {}
        # Transient failures could be costly. Retry for up to one minute.
        ret = run_command(cmd, output, num_retries=6, retry_delay_s=10)
        if ret != 0:
            logger.error(
                "Failed to run squeue command=[%s] ret=%s err=%s", cmd, ret, output["stderr"]
            )
            raise ExecutionError(f"squeue command failed: {ret}")

        return self._get_statuses_from_output(output["stdout"])

    @staticmethod
    def _get_statuses_from_output(output):
        logger.debug("squeue output:  [%s]", output)
        lines = output.split("\n")
        if not lines:
            # No jobs are currently running.
            return {}

        statuses = {}
        for line in lines:
            if line == "":
                continue
            fields = line.strip().split()
            assert len(fields) == 2
            job_id = fields[0]
            status = fields[1]
            statuses[job_id] = SlurmManager._STATUSES.get(status, HpcJobStatus.UNKNOWN)

        return statuses

    @staticmethod
    def check_storage_configuration():
        pass
        # Disabling this code because the Lustre documentation only recommends
        # higher stripe counts when files are large or if many clients will be
        # accessing the files concurrently.
        # JADE shouldn't enforce a single rule for everyone.
        # Leaving the code here in case we want to customize this in the
        # future.
        #
        # References:
        # - http://wiki.lustre.org/Configuring_Lustre_File_Striping
        # - https://www.nics.tennessee.edu/computing-resources/file-systems/lustre-striping-guide

        # output = {}
        # cmd = "lfs getstripe ."
        # ret = run_command(cmd, output)
        # if ret != 0:
        #    raise ExecutionError(f"{cmd} failed: {output}")

        # stripe_count = SlurmManager._get_stripe_count(output["stdout"])
        # logger.info("stripe_count is set to %s", stripe_count)
        # if stripe_count < 16:
        #    raise InvalidConfiguration(
        #        f"stripe_count for {os.getcwd()} is set to {stripe_count}. "
        #        "The runtime directory should be set with a stripe_count of "
        #        "16 for optimal performance. Create a new directory, run "
        #        "`lfs setstripe -c 16 <dirname>`, and then move all contents "
        #        "to that directory."
        #    )

    def get_config(self):
        return self._config

    def get_current_job_id(self):
        return os.environ["SLURM_JOB_ID"]

    @staticmethod
    def _get_stripe_count(output):
        regex = re.compile(r"stripe_count:\s+(\d+)")
        match = regex.search(output)
        assert match, output["stdout"]
        return int(match.group(1))

    def create_cluster(self):
        logger.debug("config=%s", self._config)
        assert False, "not supported"
        # cluster = SLURMCluster(
        #    project=self._config["hpc"]["allocation"],
        #    walltime=self._config["hpc"]["walltime"],
        #    job_mem=str(self._config["hpc"]["mem"]),
        #    memory=str(self._config["hpc"]["mem"]) + "MB",
        #    #job_cpu=config["cpu"],
        #    interface=self._config["dask"]["interface"],
        #    local_directory=self._config["dask"]["local_directory"],
        #    cores=self._config["dask"]["cores"],
        #    #processes=config["processes"],
        # )

        # logger.debug("Created cluster. job script %s", cluster.job_script())
        # return cluster

    def create_local_cluster(self):
        assert False, "not supported"
        # cluster = LocalCluster()
        # logger.debug("Created local cluster.")
        # return cluster

    def create_submission_script(self, name, script, filename, path):
        text = self._create_submission_script_text(name, script, path)
        utils.create_script(filename, "\n".join(text) + "\n")

    def _create_submission_script_text(self, name, script, path):
        lines = [
            "#!/bin/bash",
            f"#SBATCH --account={self._config.hpc.account}",
            f"#SBATCH --job-name={name}",
            f"#SBATCH --time={self._config.hpc.walltime}",
            f"#SBATCH --output={path}/job_output_%j.o",
            f"#SBATCH --error={path}/job_output_%j.e",
        ]

        for param in (
            "gres",
            "mem",
            "nodes",
            "ntasks",
            "ntasks_per_node",
            "partition",
            "qos",
            "tmp",
            "reservation",
        ):
            value = getattr(self._config.hpc, param, None)
            if value is not None:
                lines.append(f"#SBATCH --{param}={value}")

        lines.append("")
        lines.append(f"srun {script}")
        return lines

    def get_job_stats(self, job_id):
        cmd = (
            f"sacct -j {job_id} --format=JobID,JobName%20,state,start,end,Account,Partition%15,QOS"
        )
        output = {}
        check_run_command(cmd, output=output)
        result = output["stdout"].strip().split("\n")
        if len(result) != 6:
            raise Exception(f"Unknown output for sacct: {result} length={len(result)}")

        # 8165902       COMPLETED 2022-01-16T12:10:37 2022-01-17T04:04:34
        fields = result[2].split()
        if fields[0] != job_id:
            raise Exception(f"sacct returned unexpected job_id={fields[0]}")

        state = self._STATUSES.get(fields[2], HpcJobStatus.UNKNOWN)
        fmt = "%Y-%m-%dT%H:%M:%S"
        try:
            start = datetime.strptime(fields[3], fmt)
        except ValueError:
            logger.exception("Failed to parse start_time=%s", fields[3])
            raise
        try:
            if fields[4] == "Unknown":
                end = fields[4]
            else:
                end = datetime.strptime(fields[4], fmt)
        except ValueError:
            logger.exception("Failed to parse end_time=%s", fields[4])
            raise
        stats = HpcJobStats(
            hpc_job_id=job_id,
            name=fields[1],
            state=state,
            start=start,
            end=end,
            account=fields[5],
            partition=fields[6],
            qos=fields[7],
        )
        return stats

    def get_local_scratch(self):
        return os.environ["LOCAL_SCRATCH"]

    def get_node_id(self):
        return os.environ["SLURM_NODEID"]

    @staticmethod
    def get_num_cpus():
        return int(os.environ["SLURM_CPUS_ON_NODE"])

    def list_active_nodes(self, job_id):
        out1 = {}
        # It's possible that 500 characters won't be enough, even with the compact format.
        # Compare the node count against the result to make sure we got all nodes.
        # There should be a better way to get this.
        check_run_command(f'squeue -j {job_id} --format="%5D %500N" -h', out1)
        result = out1["stdout"].strip().split()
        assert len(result) == 2, str(result)
        num_nodes = int(result[0])
        nodes_compact = result[1]
        out2 = {}
        check_run_command(f'scontrol show hostnames "{nodes_compact}"', out2)
        nodes = [x for x in out2["stdout"].split("\n") if x != ""]
        if len(nodes) != num_nodes:
            raise Exception(f"Bug in parsing node names. Found={len(nodes)} Actual={num_nodes}")
        return nodes

    def log_environment_variables(self):
        data = {}
        for name, value in os.environ.items():
            if "SLURM" in name:
                data[name] = value

        logger.info("SLURM environment variables: %s", data)

    def submit(self, filename):
        job_id = None
        output = {}
        # Transient failures could be costly. Retry for up to one minute.
        # TODO: Some errors are not transient. We could detect those and skip the retries.
        ret = run_command("sbatch {}".format(filename), output, num_retries=6, retry_delay_s=10)
        if ret == 0:
            result = Status.GOOD
            stdout = output["stdout"]
            match = self._REGEX_SBATCH_OUTPUT.search(stdout)
            if match:
                job_id = match.group(1)
                result = Status.GOOD
            else:
                logger.error("Failed to interpret sbatch output [%s]", stdout)
                result = Status.ERROR
        else:
            result = Status.ERROR

        return result, job_id, output["stderr"]



"""Models for jobs"""

import enum
from typing import List, Optional, Set, Union

from pydantic.v1 import Field

from jade.models.base import JadeBaseModel


class JobState(enum.Enum):
    """Job states"""

    NOT_SUBMITTED = "not_submitted"
    SUBMITTED = "submitted"
    DONE = "done"


class Job(JadeBaseModel):
    """Describes the status of all jobs in the JADE config."""

    name: str = Field(
        title="name",
        description="name of the job, must be unique",
    )
    blocked_by: Set[str] = Field(
        title="blocked_by",
        description="job names that are blocking this job",
    )
    cancel_on_blocking_job_failure: bool = Field(
        title="cancel_on_blocking_job_failure",
        description="cancel job if one of its blocking jobs fails",
        default=False,
    )
    state: JobState = Field(
        title="state",
        description="job state",
    )


class JobStatus(JadeBaseModel):
    """Describes all jobs in the JADE config."""

    jobs: List[Job] = Field(
        title="jobs",
        description="describes all jobs in the JADE config",
    )
    hpc_job_ids: List[str] = Field(
        title="hpc_job_ids",
        description="HPC job IDs for active jobs",
    )
    batch_index: int = Field(
        title="batch_index",
        description="HPC batch index",
        default=1,
    )
    version: int = Field(
        title="version",
        description="version of the statuses, increments with each update",
    )



"""CLI to show extensions."""

import click

from jade.extensions.registry import Registry
from jade.loggers import setup_logging
from jade.utils.utils import load_data


@click.group()
def extensions():
    """Manage JADE extensions."""
    setup_logging("extensions", None)


@click.command()
@click.argument("extension-file")
def register(extension_file):
    """Register one or more extensions."""
    registry = Registry()
    for extension in load_data(extension_file):
        registry.register_extension(extension)


@click.command()
def reset_defaults():
    """Reset registry to its default values."""
    Registry().reset_defaults()


@click.command()
@click.argument("extension")
def unregister(extension):
    """Unregister an extension."""
    registry = Registry()
    registry.unregister_extension(extension)


@click.command()
@click.argument("package-name")
def add_logger(package_name):
    """Add logging for a package."""
    registry = Registry()
    registry.add_logger(package_name)


@click.command()
@click.argument("package-name")
def remove_logger(package_name):
    """Remove logging for a package."""
    registry = Registry()
    registry.remove_logger(package_name)


@click.command()
def show():
    """Show the available extensions (job types)."""
    print("Extensions:")
    Registry().show_extensions()
    print("Logging enabled for packages:  ", end="")
    Registry().show_loggers()


extensions.add_command(register)
extensions.add_command(reset_defaults)
extensions.add_command(unregister)
extensions.add_command(add_logger)
extensions.add_command(remove_logger)
extensions.add_command(show)



"""CLI to run jobs."""

import logging
import os
import shutil
import sys

import click

from jade.common import OUTPUT_DIR
from jade.cli.common import COMMON_SUBMITTER_OPTIONS, add_options, make_submitter_params
from jade.jobs.job_configuration_factory import create_config_from_file
from jade.jobs.job_submitter import JobSubmitter
from jade.loggers import setup_logging, setup_event_logging
from jade.hpc.common import HpcType
from jade.models.hpc import SlurmConfig
from jade.models.submitter_params import SubmitterParams
from jade.jobs.cluster import Cluster
from jade.utils.utils import get_cli_string, load_data


@click.command()
@click.argument(
    "config-file",
    type=str,
)
@click.option(
    "-f",
    "--force",
    default=False,
    is_flag=True,
    show_default=True,
    help="Delete output directory if it exists.",
)
@click.option("-o", "--output", default=OUTPUT_DIR, show_default=True, help="Output directory.")
@click.option(
    "-s",
    "--submitter-params",
    default=None,
    show_default=False,
    type=str,
    help="Filename with submitter parameters. Supercedes other CLI parameters. Generate defaults "
    "with 'jade config submitter-params'",
)
@add_options(COMMON_SUBMITTER_OPTIONS)
def submit_jobs(
    config_file=None,
    per_node_batch_size=None,
    dry_run=None,
    force=None,
    hpc_config=None,
    local=None,
    max_nodes=None,
    output=None,
    poll_interval=None,
    resource_monitor_interval=None,
    resource_monitor_type=None,
    resource_monitor_stats=None,
    num_parallel_processes_per_node=None,
    verbose=None,
    reports=None,
    enable_singularity=None,
    container=None,
    try_add_blocked_jobs=None,
    time_based_batching=None,
    node_setup_script=None,
    node_shutdown_script=None,
    submitter_params=None,
    no_distributed_submitter=None,
):
    """Submits jobs for execution, locally or on HPC."""
    if os.path.exists(output):
        if force:
            shutil.rmtree(output)
        else:
            print(
                f"{output} already exists. Delete it or use '--force' to overwrite.",
                file=sys.stderr,
            )
            sys.exit(1)

    os.makedirs(output)
    filename = os.path.join(output, "submit_jobs.log")
    event_filename = os.path.join(output, "submit_jobs_events.log")
    level = logging.DEBUG if verbose else logging.INFO
    # For some reason event logging must be setup before general logging.
    # Otherwise, the first event doesn't show up in the log.
    setup_event_logging(event_filename, mode="a")
    logger = setup_logging(__name__, filename, file_level=level, console_level=level, mode="a")
    logger.info(get_cli_string())

    config = create_config_from_file(config_file)
    if config.submission_groups:
        params = config.submission_groups[0].submitter_params
    elif submitter_params is not None:
        params = SubmitterParams(**load_data(submitter_params))
    else:
        params = make_submitter_params(
            per_node_batch_size=per_node_batch_size,
            dry_run=dry_run,
            hpc_config=hpc_config,
            local=local,
            max_nodes=max_nodes,
            poll_interval=poll_interval,
            resource_monitor_interval=resource_monitor_interval,
            resource_monitor_type=resource_monitor_type,
            resource_monitor_stats=resource_monitor_stats,
            num_parallel_processes_per_node=num_parallel_processes_per_node,
            verbose=verbose,
            reports=reports,
            enable_singularity=enable_singularity,
            container=container,
            try_add_blocked_jobs=try_add_blocked_jobs,
            time_based_batching=time_based_batching,
            node_setup_script=node_setup_script,
            node_shutdown_script=node_shutdown_script,
            no_distributed_submitter=no_distributed_submitter,
        )

    if not config.submission_groups:
        config.assign_default_submission_group(params)

    for group in config.submission_groups:
        params = group.submitter_params
        if params.resource_monitor_interval is not None:
            if params.resource_monitor_interval < params.poll_interval:
                logger.warning(
                    "resource_monitor_interval cannot be less than poll_interval. "
                    "Reducing poll_interval"
                )
                params.poll_interval = params.resource_monitor_interval

    try:
        ret = JobSubmitter.run_submit_jobs(
            config,
            output,
            local=params.hpc_config.hpc_type == HpcType.LOCAL,
            dry_run=params.dry_run,
        )
        sys.exit(ret)
    except Exception:
        logger.exception("Failed to run submit_jobs")
        raise
    finally:
        logging.shutdown()



from pydantic.v1 import BaseModel

from jade.models.base import JadeBaseModel
from jade.models.hpc import HpcConfig, SlurmConfig, FakeHpcConfig, LocalHpcConfig
from jade.models.singularity import SingularityParams
from jade.models.submitter_params import SubmitterParams
from jade.models.submission_group import SubmissionGroup
from jade.models.jobs import Job, JobState, JobStatus
from jade.models.cluster_config import ClusterConfig
from jade.models.pipeline import PipelineConfig, PipelineStage


def get_model_defaults(model_class: BaseModel):
    """Return the default values for fields in a Pydantic BaseModel.
    If a field doesn't have a default then return None.
    Default values may also be None.

    Returns
    -------
    dict

    """
    return {
        x: y.get("default") for x, y in model_class.schema(by_alias=False)["properties"].items()
    }



"""Defines interface for simulations."""


import abc
import logging


logger = logging.getLogger(__name__)


class JobExecutionInterface(abc.ABC):
    """Interface definition for simulations."""

    @staticmethod
    def collect_results(output_dir):
        """Collect result data from output directory.

        Parameters
        ----------
        output_dir : str

        Returns
        -------
        list of dict

        """
        logger.debug("Collect results from %s", output_dir)
        # subclasses can override.
        return []

    @classmethod
    @abc.abstractmethod
    def create(cls, job_inputs, job, output):
        """Creates an instance of a JobExecutionInterface."""

    @staticmethod
    @abc.abstractmethod
    def generate_command(job, output, config_file, verbose=False):
        """Generate a command for a job to be run in a subprocess.

        Parameters
        ----------
        job : class.parameters_type
        output : str
            output directory
        config_file : str
            job configuration file

        Returns
        -------
        str
            command that can be executed in the OS

        """

    @property
    @abc.abstractmethod
    def results_directory(self):
        """Return the results directory created by the simulation."""

    @abc.abstractmethod
    def list_results_files(self):
        """Return a list of result filenames created by the simulation."""

    @abc.abstractmethod
    def post_process(self, **kwargs):
        """Run post-process operations on data."""

    @abc.abstractmethod
    def run(self):
        """Runs the simulation."""



"""The job execution class and methods for generic_command."""

import logging
import os

from jade.jobs.job_execution_interface import JobExecutionInterface


logger = logging.getLogger(__name__)


class GenericCommandExecution(JobExecutionInterface):
    """A class used for executing a generic_command."""

    def __init__(self, job, output):
        """Init generic_command execution class

        Parameters
        ----------
        job: :obj:`GenericCommandParameters`
            The instance of :obj:`GenericCommandParameters`
        output: str,
            The path to the output directory.

        """
        self._job = job
        self._output = output

    @property
    def results_directory(self):
        return self._output

    @classmethod
    def create(cls, _, job, output):
        return cls(job, output)

    @staticmethod
    def generate_command(job, output, config_file, verbose=False):
        # These jobs already have a command and are not run with jade-internal.
        cmd = job.command
        if job.append_job_name:
            cmd += f" --jade-job-name={job.name}"
        if job.append_output_dir:
            # output is jobs-output
            output_dir = os.path.dirname(output)
            cmd += f" --jade-runtime-output={output_dir}"
        return cmd

    def list_results_files(self):
        return []

    def post_process(self, **kwargs):
        pass

    def run(self):
        assert False



"""
This module contains StructuredLogEvent and EventSummary classes.
"""

from collections import defaultdict
import json
import logging
import os
import sys
from datetime import datetime
from pathlib import Path

import pandas as pd
from prettytable import PrettyTable

from jade.common import JOBS_OUTPUT_DIR, EVENTS_DIR
from jade.exceptions import InvalidConfiguration
from jade.utils.utils import dump_data, load_data


EVENTS_FILENAME = "events.json"

EVENT_CATEGORY_ERROR = "Error"
EVENT_CATEGORY_HPC = "HPC"
EVENT_CATEGORY_RESOURCE_UTIL = "ResourceUtilization"

EVENT_NAME_HPC_SUBMIT = "hpc_submit"
EVENT_NAME_HPC_JOB_ASSIGNED = "hpc_job_assigned"
EVENT_NAME_HPC_JOB_STATE_CHANGE = "hpc_job_state_change"
EVENT_NAME_CPU_STATS = "cpu_stats"
EVENT_NAME_DISK_STATS = "disk_stats"
EVENT_NAME_MEMORY_STATS = "mem_stats"
EVENT_NAME_NETWORK_STATS = "net_stats"
EVENT_NAME_PROCESS_STATS = "process_stats"
EVENT_NAME_BYTES_CONSUMED = "bytes_consumed"
EVENT_NAME_UNHANDLED_ERROR = "unhandled_error"
EVENT_NAME_ERROR_LOG = "log_error"
EVENT_NAME_SUBMIT_STARTED = "submit_started"
EVENT_NAME_SUBMIT_COMPLETED = "submit_completed"
EVENT_NAME_CONFIG_EXEC_SUMMARY = "config_exec_summary"

logger = logging.getLogger(__name__)


class StructuredLogEvent:
    """
    A class for recording structured log events.
    """

    def __init__(self, source, category, name, message, **kwargs):
        """
        Initialize the class

        Parameters
        ----------
        source: str,
            The source of the event.
        category: str,
            An event category given by the user.
        name: str,
            An event name given by the user.
        message:
            An event message given the user.

        kwargs:
            Other information that the user needs to record into event.
        """
        self.source = source
        self.category = category
        self.name = name
        self.message = message
        self.event_class = self.__class__.__name__

        if "timestamp" in kwargs:
            self.timestamp = kwargs.pop("timestamp")
        else:
            self.timestamp = str(datetime.now())

        self.data = kwargs

    def base_field_names(self):
        """Return the base field names for the event.

        Returns
        -------
        list

        """
        return self._base_field_names()

    @staticmethod
    def _base_field_names():
        return ["timestamp", "source", "message"]

    def field_names(self):
        """Return all field names for the event.

        Returns
        -------
        list

        """
        return self.base_field_names() + list(self.data.keys())

    def values(self):
        """Return the values for all fields in the event.

        Returns
        -------
        list

        """
        # Account for events generated with different versions of code.
        values = [getattr(self, x, "") for x in self.base_field_names()]
        values += [self.data.get(x, "") for x in self.data]
        return values

    @classmethod
    def deserialize(cls, record):
        """Deserialize event from JSON.

        Parameters
        ----------
        record : dict

        Returns
        -------
        StructuredLogEvent

        """
        return cls(
            source=record.get("source", ""),
            category=record.get("category", ""),
            name=record.get("name", ""),
            message=record.get("message", ""),
            timestamp=record.get("timestamp", ""),
            **record["data"],
        )

    def __str__(self):
        """To format a event instance to string"""
        return json.dumps(self.__dict__, sort_keys=True)

    def to_dict(self):
        """Convert event object to dict"""
        return self.__dict__


class StructuredErrorLogEvent(StructuredLogEvent):
    """Event specific to exceptions"""

    def __init__(self, source, category, name, message, **kwargs):
        """Must be called in an exception context."""
        super().__init__(source, category, name, message, **kwargs)
        if "exception" not in kwargs:
            self._parse_traceback()

    def base_field_names(self):
        return self._base_field_names()

    def _parse_traceback(self):
        """
        Parse the system exception information - exception, filename, and lineno.
        """
        exc_type, exc_obj, tb = sys.exc_info()
        assert tb is not None, "must be called in an exception context"

        self.data["exception"] = str(exc_type)
        self.data["error"] = str(exc_obj).strip()
        self.data["filename"] = os.path.basename(tb.tb_frame.f_code.co_filename)
        self.data["lineno"] = tb.tb_lineno


def deserialize_event(data):
    """Construct an event from raw  data.

    Parameters
    ----------
    data : dict

    Returns
    -------
    StructuredLogEvent

    """
    if data["event_class"] == "StructuredLogEvent":
        return StructuredLogEvent.deserialize(data)
    if data["event_class"] == "StructuredErrorLogEvent":
        return StructuredErrorLogEvent.deserialize(data)

    raise Exception(f"unknown event class {data['event_class']}")


class EventsSummary:
    """Provides summary of all events."""

    RESOURCE_STATS = set(
        (
            EVENT_NAME_CPU_STATS,
            EVENT_NAME_DISK_STATS,
            EVENT_NAME_MEMORY_STATS,
            EVENT_NAME_NETWORK_STATS,
            EVENT_NAME_PROCESS_STATS,
        )
    )

    def __init__(self, output_dir, preload=False, optimize_resoure_stats=True):
        """
        Initialize EventsSummary class

        Parameters
        ----------
        output_dir: str
            Path of jade output directory.
        preload: bool
            Load all events into memory; otherwise, load by name on demand.

        """
        self._optimize_resource_stats = optimize_resoure_stats
        self._events = defaultdict(list)
        self._output_dir = output_dir
        self._event_dir = Path(output_dir) / EVENTS_DIR
        self._event_dir.mkdir(exist_ok=True)
        self._job_outputs_dir = os.path.join(output_dir, JOBS_OUTPUT_DIR)
        event_files = list(self._event_dir.iterdir())
        if not event_files:
            self._consolidate_events()
            self._save_events_summary()
        elif preload:
            self._load_all_events()
        # else, events have already been consolidated, load them on demand

    def _iter_event_files(self):
        return Path(self._output_dir).glob("*events.log")

    def _consolidate_events(self):
        """Find most recent event log files, and merge event data together."""
        # PERF: This could easily be parallelized across processes. Need to be sure we aren't on
        # a login node.
        for event_file in self._iter_event_files():
            with open(event_file, "r") as f:
                for line in f:
                    record = json.loads(line)
                    event = deserialize_event(record)
                    self._events[event.name].append(event)

        for name in self._events.keys():
            self._events[name].sort(key=lambda x: x.timestamp)

    def _deserialize_events(self, name, path):
        self._events[name] = [deserialize_event(x) for x in load_data(path)]

    def _get_events(self, name):
        if name not in self._events:
            self._load_event_file(name)
        return self._events.get(name, [])

    def _load_all_events(self):
        for filename in self._event_dir.iterdir():
            name = filename.stem
            if name not in self._events and name not in self.RESOURCE_STATS:
                self._deserialize_events(name, filename)

    def _load_event_file(self, name):
        if name in self.RESOURCE_STATS:
            raise Exception(f"event {name} is only available in Parquet")

        filename = self._make_event_filename(name)
        if filename.exists():
            self._deserialize_events(name, filename)

    def _make_data_filename(self, name):
        return self._event_dir / (name + ".parquet")

    def _make_event_filename(self, name):
        return self._event_dir / (name + ".json")

    def _save_events_summary(self):
        """Save events to one file per event name."""
        for name, events in self._events.items():
            if name in self.RESOURCE_STATS:
                dict_events = []
                for event in events:
                    data = {"timestamp": event.timestamp, "source": event.source}
                    if name == EVENT_NAME_PROCESS_STATS:
                        for process in event.data["processes"]:
                            _data = {"timestamp": event.timestamp, "source": event.source}
                            _data.update(process)
                            dict_events.append(_data)
                    else:
                        data.update(event.data)
                        dict_events.append(data)
                df = pd.DataFrame.from_records(dict_events, index="timestamp")
                filename = self._make_data_filename(name)
                df.to_parquet(filename)
            else:
                dict_events = [event.to_dict() for event in events]
                filename = self._make_event_filename(name)
                dump_data(dict_events, filename)

        for name in self.RESOURCE_STATS:
            self._events.pop(name, None)

    def get_bytes_consumed(self):
        """Return a sum of all bytes_consumed events.

        Returns
        -------
        int
            Size in bytes of files produced by all jobs

        """
        total = 0
        for event in self.iter_events(EVENT_NAME_BYTES_CONSUMED):
            total += event.data["bytes_consumed"]

        return total

    def get_config_exec_time(self):
        """Return the total number of seconds to run all jobs in the config.

        Returns
        -------
        int

        """
        events = self.list_events(EVENT_NAME_CONFIG_EXEC_SUMMARY)
        if not events:
            raise InvalidConfiguration("no batch summary events found")

        return events[0].data["config_execution_time"]

    def get_dataframe(self, name):
        """Return the dataframe for this event name. Only applicable to resource stats."""
        filename = self._make_data_filename(name)
        if not filename.exists():
            return pd.DataFrame()
        return pd.read_parquet(filename)

    def iter_events(self, name):
        """Return a generator over events with name.

        Parameters
        ----------
        name : str

        Yields
        ------
        event : StructuredLogEvent

        """
        for event in self._get_events(name):
            yield event

    def list_events(self, name):
        """Return the events of type name.

        Returns
        -------
        list
            list of StructuredLogEvent

        """
        return list(self._get_events(name))

    def list_unique_categories(self):
        """Return the unique event categories in the log. Will cause all events
        to get loaded into memory.

        Returns
        -------
        list

        """
        self._load_all_events()
        categories = set()
        for events in self._events.values():
            if not events:
                continue
            categories.add(events[0].category)

        categories = list(categories)
        categories.sort()
        return categories

    def list_unique_names(self):
        """Return the unique event names in the log.

        Returns
        -------
        list

        """
        return [x.name for x in self._event_dir.iterdir() if x.suffix == ".json"]

    def show_events(self, name):
        """Print tabular events in terminal"""
        table = PrettyTable()

        field_names = None
        count = 0
        for event in self.iter_events(name):
            if field_names is None:
                field_names = event.field_names()
            table.add_row(event.values())
            count += 1

        if count == 0:
            print(f"No events of type {name}")
            return

        table.field_names = field_names
        print(f"Events of type {name} from directory: {self._output_dir}")
        print(table)
        print(f"Total events: {count}\n")

    def show_events_in_category(self, category):
        """Print tabular events matching category in terminal. Will cause all
        events to get loaded into memory.

        """
        event_names = []
        self._load_all_events()
        for name, events in self._events.items():
            if not events:
                continue
            if events[0].category == category:
                event_names.append(name)

        if not event_names:
            print(f"There are no events in category {category}")
            return

        for event_name in sorted(event_names):
            self.show_events(event_name)

    def show_event_categories(self):
        """Show the unique event categories in the log."""
        print("Catgories:  {}".format(" ".join(self.list_unique_categories())))

    def show_event_names(self):
        """Show the unique event names in the log."""
        print("Names:  {}".format(" ".join(self.list_unique_names())))

    def to_json(self):
        """Return all events in JSON format.

        Returns
        -------
        str

        """
        self._load_all_events()
        return json.dumps(
            [x.to_dict() for events in self._events.values() for x in events], indent=2
        )



"""
The job configuration class for collecting all jobs.
"""
import os
from jade.jobs.job_inputs_interface import JobInputsInterface
from jade.utils.utils import load_data
from jade.extensions.demo.autoregression_parameters import AutoRegressionParameters


class AutoRegressionInputs(JobInputsInterface):
    """
    A class used for configuring auto-regression analysis jobs.
    """

    INPUTS_FILE = "inputs.json"

    def __init__(self, base_directory):
        self._base_directory = base_directory
        self._parameters = {}
        self.get_available_parameters()

    @property
    def base_directory(self):
        """The directory contains inputs data"""
        return self._base_directory

    def get_available_parameters(self):
        """Collect all available auto-regression jobs"""
        inputs_file = os.path.join(self._base_directory, self.INPUTS_FILE)
        inputs = load_data(inputs_file)

        for param in inputs:
            job = AutoRegressionParameters(
                country=param["country"], data=os.path.join(self._base_directory, param["data"])
            )

            self._parameters[job.name] = job

    def iter_jobs(self):
        """Return a list of auto-regression jobs"""
        return list(self._parameters.values())



"""Defines parameters for submitting jobs to an HPC."""

from pydantic.v1 import Field

from jade.models import JadeBaseModel, SubmitterParams


class SubmissionGroup(JadeBaseModel):
    """Defines parameters for submitting jobs to an HPC."""

    name: str = Field(
        title="name",
        description="User-defined name of the group",
    )
    submitter_params: SubmitterParams = Field(
        title="submitter_params",
        description="Submission parameters for the group",
    )


def make_submission_group_lookup(submission_groups):
    """Return the submission groups in a dict keyed by name.

    Returns
    -------
    dict

    """
    return {x.name: x for x in submission_groups}



import abc
import logging
import sys
import time
from collections import defaultdict
from pathlib import Path

import pandas as pd
import plotly.graph_objects as go
from prettytable import PrettyTable
import psutil
from psutil._common import bytes2human
from tabulate import tabulate

from jade.common import STATS_DIR
from jade.events import (
    EVENT_CATEGORY_RESOURCE_UTIL,
    EVENT_NAME_CPU_STATS,
    EVENT_NAME_DISK_STATS,
    EVENT_NAME_MEMORY_STATS,
    EVENT_NAME_NETWORK_STATS,
    EVENT_NAME_PROCESS_STATS,
    StructuredLogEvent,
)
from jade.loggers import log_event
from jade.models.submitter_params import ResourceMonitorStats
from jade.utils.utils import dump_data


logger = logging.getLogger(__name__)

ONE_MB = 1024 * 1024


class ResourceMonitor:
    """Monitors resource utilization statistics"""

    DISK_STATS = (
        "read_count",
        "write_count",
        "read_bytes",
        "write_bytes",
        "read_time",
        "write_time",
    )
    NET_STATS = (
        "bytes_recv",
        "bytes_sent",
        "dropin",
        "dropout",
        "errin",
        "errout",
        "packets_recv",
        "packets_sent",
    )

    def __init__(self, name):
        self._name = name
        self._last_disk_check_time = None
        self._last_net_check_time = None
        self._update_disk_stats(psutil.disk_io_counters())
        self._update_net_stats(psutil.net_io_counters())
        self._cached_processes = {}  # pid to psutil.Process

    def _update_disk_stats(self, data):
        for stat in self.DISK_STATS:
            setattr(self, stat, getattr(data, stat, 0))
        self._last_disk_check_time = time.time()

    def _update_net_stats(self, data):
        for stat in self.NET_STATS:
            setattr(self, stat, getattr(data, stat, 0))
        self._last_net_check_time = time.time()

    def get_cpu_stats(self):
        """Gets CPU current resource stats information."""
        stats = psutil.cpu_times_percent()._asdict()
        stats["cpu_percent"] = psutil.cpu_percent()
        return stats

    def get_disk_stats(self):
        """Gets current disk stats."""
        data = psutil.disk_io_counters()
        stats = {
            "elapsed_seconds": time.time() - self._last_disk_check_time,
        }
        for stat in self.DISK_STATS:
            stats[stat] = getattr(data, stat, 0) - getattr(self, stat, 0)
        stats["read MB/s"] = self._mb_per_sec(stats["read_bytes"], stats["elapsed_seconds"])
        stats["write MB/s"] = self._mb_per_sec(stats["write_bytes"], stats["elapsed_seconds"])
        stats["read IOPS"] = float(stats["read_count"]) / stats["elapsed_seconds"]
        stats["write IOPS"] = float(stats["write_count"]) / stats["elapsed_seconds"]
        self._update_disk_stats(data)
        return stats

    def get_memory_stats(self):
        """Gets current memory resource stats."""
        return psutil.virtual_memory()._asdict()

    def get_network_stats(self):
        """Gets current network stats."""
        data = psutil.net_io_counters()
        stats = {
            "elapsed_seconds": time.time() - self._last_net_check_time,
        }
        for stat in self.NET_STATS:
            stats[stat] = getattr(data, stat, 0) - getattr(self, stat, 0)
        stats["recv MB/s"] = self._mb_per_sec(stats["bytes_recv"], stats["elapsed_seconds"])
        stats["sent MB/s"] = self._mb_per_sec(stats["bytes_sent"], stats["elapsed_seconds"])
        self._update_net_stats(data)
        return stats

    def _get_process(self, pid):
        process = self._cached_processes.get(pid)
        if process is None:
            try:
                process = psutil.Process(pid)
                # Initialize CPU utilization tracking per psutil docs.
                process.cpu_percent(interval=0.2)
                self._cached_processes[pid] = process
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                logger.debug("Tried to construct Process for invalid pid=%s", pid)
                return None

        return process

    def clear_stale_processes(self, cur_pids):
        """Remove cached process objects that are no longer running."""
        self._cached_processes = {
            pid: proc for pid, proc in self._cached_processes.items() if pid in cur_pids
        }

    def get_process_stats(self, pid, include_children=True, recurse_children=False):
        """Gets current process stats. Returns None if the pid does not exist."""
        children = []
        process = self._get_process(pid)
        if process is None:
            return None, children
        try:
            with process.oneshot():
                stats = {
                    "rss": process.memory_info().rss,
                    "cpu_percent": process.cpu_percent(),
                }
                if include_children:
                    for child in process.children(recursive=recurse_children):
                        cached_child = self._get_process(child.pid)
                        if cached_child is not None:
                            stats["cpu_percent"] += cached_child.cpu_percent()
                            stats["rss"] += cached_child.memory_info().rss
                            children.append(child.pid)
                return stats, children
        except (psutil.NoSuchProcess, psutil.AccessDenied):
            logger.debug("Tried to get process info for invalid pid=%s", pid)
            return None, []

    @staticmethod
    def _mb_per_sec(num_bytes, elapsed_seconds):
        return float(num_bytes) / ONE_MB / elapsed_seconds

    @property
    def name(self):
        """Return the name of the monitor."""
        return self._name


class ResourceMonitorAggregator:
    """Aggregates resource utilization stats in memory."""

    def __init__(
        self,
        name,
        stats: ResourceMonitorStats,
    ):
        self._stats = stats
        self._count = 0
        self._monitor = ResourceMonitor(name)
        self._last_stats = self._get_stats()
        self._summaries = {
            "average": defaultdict(dict),
            "maximum": defaultdict(dict),
            "minimum": defaultdict(dict),
            "sum": defaultdict(dict),
        }
        for resource_type, stat_dict in self._last_stats.items():
            for stat_name in stat_dict:
                self._summaries["average"][resource_type][stat_name] = 0.0
                self._summaries["maximum"][resource_type][stat_name] = 0.0
                self._summaries["minimum"][resource_type][stat_name] = sys.maxsize
                self._summaries["sum"][resource_type][stat_name] = 0.0

        self._process_summaries = {
            "average": defaultdict(dict),
            "maximum": defaultdict(dict),
            "minimum": defaultdict(dict),
            "sum": defaultdict(dict),
        }
        self._process_sample_count = {}

    def _get_stats(self):
        data = {}
        if self._stats.cpu:
            data[CpuStatsViewer.metric()] = self._monitor.get_cpu_stats()
        if self._stats.disk:
            data[DiskStatsViewer.metric()] = self._monitor.get_disk_stats()
        if self._stats.memory:
            data[MemoryStatsViewer.metric()] = self._monitor.get_memory_stats()
        if self._stats.network:
            data[NetworkStatsViewer.metric()] = self._monitor.get_network_stats()
        return data

    def _get_process_stats(self, pids):
        stats = {}
        cur_pids = set()
        for name, pid in pids.items():
            _stats, children = self._monitor.get_process_stats(
                pid,
                include_children=self._stats.include_child_processes,
                recurse_children=self._stats.recurse_child_processes,
            )
            if _stats is not None:
                stats[name] = _stats
                cur_pids.add(pid)
                for child in children:
                    cur_pids.add(child)

        self._monitor.clear_stale_processes(cur_pids)
        return stats

    def finalize(self, output_dir):
        """Finalize the stat summaries and record the results.

        Parameters
        ----------
        output_dir : str
            Directory in which to record the results.

        """
        if self._count == 0:
            logger.info("Resource monitoring was disabled")
            return

        for resource_type, stat_dict in self._summaries["sum"].items():
            for stat_name, val in stat_dict.items():
                self._summaries["average"][resource_type][stat_name] = val / self._count

        self._summaries.pop("sum")
        stat_summaries = []
        for resource_type in (
            CpuStatsViewer.metric(),
            DiskStatsViewer.metric(),
            MemoryStatsViewer.metric(),
            NetworkStatsViewer.metric(),
        ):
            # Make each entry look like what the stat viewers produce.
            summary = {"batch": self.name, "type": resource_type}
            for stat_type in self._summaries.keys():
                summary[stat_type] = self._summaries[stat_type][resource_type]
            stat_summaries.append(summary)

        for process_name, stat_dict in self._process_summaries["sum"].items():
            for stat_name, val in stat_dict.items():
                self._process_summaries["average"][process_name][stat_name] = (
                    val / self._process_sample_count[process_name]
                )

        self._process_summaries.pop("sum")
        for process_name, samples in self._process_sample_count.items():
            summary = {
                "batch": self.name,
                "name": process_name,
                "samples": samples,
                "type": ProcessStatsViewer.metric(),
            }
            for stat_type in self._process_summaries.keys():
                summary[stat_type] = self._process_summaries[stat_type][process_name]
            stat_summaries.append(summary)

        path = Path(output_dir) / STATS_DIR
        filename = path / f"{self.name}_resource_stats.json"
        dump_data(stat_summaries, filename)

    @property
    def name(self):
        """Return the name of the monitor."""
        return self._monitor.name

    def update_resource_stats(self, ids=None):
        """Update resource stats information as structured job events for the current interval."""
        cur_stats = self._get_stats()
        for resource_type, stat_dict in cur_stats.items():
            for stat_name, val in stat_dict.items():
                if val > self._summaries["maximum"][resource_type][stat_name]:
                    self._summaries["maximum"][resource_type][stat_name] = val
                elif val < self._summaries["minimum"][resource_type][stat_name]:
                    self._summaries["minimum"][resource_type][stat_name] = val
                self._summaries["sum"][resource_type][stat_name] += val

        if self._stats.process:
            cur_process_stats = self._get_process_stats(ids)
            for process_name, stat_dict in cur_process_stats.items():
                if process_name in self._process_summaries["maximum"]:
                    for stat_name, val in stat_dict.items():
                        if val > self._process_summaries["maximum"][process_name][stat_name]:
                            self._process_summaries["maximum"][process_name][stat_name] = val
                        elif val < self._process_summaries["minimum"][process_name][stat_name]:
                            self._process_summaries["minimum"][process_name][stat_name] = val
                        self._process_summaries["sum"][process_name][stat_name] += val
                    self._process_sample_count[process_name] += 1
                else:
                    for stat_name, val in stat_dict.items():
                        self._process_summaries["maximum"][process_name][stat_name] = val
                        self._process_summaries["minimum"][process_name][stat_name] = val
                        self._process_summaries["sum"][process_name][stat_name] = val
                    self._process_sample_count[process_name] = 1

        self._count += 1
        self._last_stats = cur_stats


class ResourceMonitorLogger:
    """Logs resource utilization stats on periodic basis."""

    def __init__(
        self,
        name,
        stats: ResourceMonitorStats,
        include_child_processes=True,
        recurse_child_processes=False,
    ):
        self._monitor = ResourceMonitor(name)
        self._stats = stats
        self._include_child_processes = include_child_processes
        self._recurse_child_processes = recurse_child_processes

    def log_cpu_stats(self):
        """Logs CPU resource stats information."""
        cpu_stats = self._monitor.get_cpu_stats()
        log_event(
            StructuredLogEvent(
                source=self.name,
                category=EVENT_CATEGORY_RESOURCE_UTIL,
                name=EVENT_NAME_CPU_STATS,
                message="Node CPU stats update",
                **cpu_stats,
            )
        )

    def log_disk_stats(self):
        """Logs disk stats."""
        stats = self._monitor.get_disk_stats()
        log_event(
            StructuredLogEvent(
                source=self.name,
                category=EVENT_CATEGORY_RESOURCE_UTIL,
                name=EVENT_NAME_DISK_STATS,
                message="Node disk stats update",
                **stats,
            )
        )

    def log_memory_stats(self):
        """Logs memory resource stats information."""
        mem_stats = self._monitor.get_memory_stats()
        log_event(
            StructuredLogEvent(
                source=self.name,
                category=EVENT_CATEGORY_RESOURCE_UTIL,
                name=EVENT_NAME_MEMORY_STATS,
                message="Node memory stats update",
                **mem_stats,
            )
        )

    def log_network_stats(self):
        """Logs memory resource stats information."""
        stats = self._monitor.get_network_stats()
        log_event(
            StructuredLogEvent(
                source=self.name,
                category=EVENT_CATEGORY_RESOURCE_UTIL,
                name=EVENT_NAME_NETWORK_STATS,
                message="Node net stats update",
                **stats,
            )
        )

    def log_process_stats(self, pids):
        """Log stats for each process.

        Parameters
        ----------
        ids : dict, defaults to None
            Maps job name to process ID.

        """
        stats = {"processes": []}
        cur_pids = set()
        for name, pid in pids.items():
            stat, children = self._monitor.get_process_stats(
                pid,
                include_children=self._stats.include_child_processes,
                recurse_children=self._stats.recurse_child_processes,
            )
            if stat is not None:  # The process could have exited.
                stat["name"] = name
                stats["processes"].append(stat)
                cur_pids.add(pid)
                for child in children:
                    cur_pids.add(child)

        self._monitor.clear_stale_processes(cur_pids)

        if stats["processes"]:
            log_event(
                StructuredLogEvent(
                    source=self.name,
                    category=EVENT_CATEGORY_RESOURCE_UTIL,
                    name=EVENT_NAME_PROCESS_STATS,
                    message="Process stats update",
                    **stats,
                )
            )

    def log_resource_stats(self, ids=None):
        """Logs resource stats information as structured job events for the current interval.

        Parameters
        ----------
        ids : dict, defaults to None
            Maps job name to process ID.

        """
        if self._stats.cpu:
            self.log_cpu_stats()
        if self._stats.disk:
            self.log_disk_stats()
        if self._stats.memory:
            self.log_memory_stats()
        if self._stats.network:
            self.log_network_stats()
        if self._stats.process and ids is not None:
            self.log_process_stats(ids)

    @property
    def name(self):
        """Return the name of the monitor."""
        return self._monitor.name


class StatsViewerBase(abc.ABC):
    """Base class for viewing statistics"""

    def __init__(self, events, event_name):
        self._event_name = event_name
        self._df_by_batch = {}

        df = events.get_dataframe(event_name)
        if not df.empty:
            for batch in df["source"].unique():
                self._df_by_batch[batch] = df.loc[df["source"] == batch]

    @staticmethod
    def get_printable_value(field, val):
        if isinstance(val, float):
            return "{:.3f}".format(val)
        return str(val)

    def get_dataframe(self, batch):
        """Return a dataframe for the batch's stats.

        Parameters
        ----------
        batch : str

        Returns
        -------
        pd.DataFrame

        """
        if batch not in self._df_by_batch:
            return pd.DataFrame()
        return self._df_by_batch[batch]

    def iter_batch_names(self):
        """Return an iterator over the batch names."""
        return self._df_by_batch.keys()

    def plot_to_file(self, output_dir):
        """Make plots of resource utilization for one node.

        Parameters
        ----------
        directory : str
            output directory

        """
        if pd.options.plotting.backend != "plotly":
            pd.options.plotting.backend = "plotly"

        exclude = self._non_plottable_columns()
        for name in self.iter_batch_names():
            df = self.get_dataframe(name)
            cols = [x for x in df.columns if x not in exclude]
            title = f"{self.__class__.__name__} {name}"
            fig = df[cols].plot(title=title)
            filename = Path(output_dir) / f"{self.__class__.__name__}__{name}.html"
            fig.write_html(str(filename))
            logger.info("Generated plot in %s", filename)

    @staticmethod
    @abc.abstractmethod
    def metric():
        """Return the metric."""

    @staticmethod
    def _non_plottable_columns():
        """Return the columns that cannot be plotted."""
        return {"source"}

    def show_stats(self, show_all_timestamps=True):
        """Show statistics"""
        text = f"{self.metric()} statistics for each batch"
        print(f"\n{text}")
        print("=" * len(text) + "\n")
        self._show_stats(show_all_timestamps=show_all_timestamps)

    def get_stats_summary(self):
        """Return a list of objects describing summaries of all stats.

        Returns
        -------
        list
            list of dicts

        """
        stats = []
        for batch, df in self._df_by_batch.items():
            if df.empty:
                continue
            entry = {
                "type": self.metric(),
                "batch": batch,
                "average": {},
                "minimum": {},
                "maximum": {},
            }
            exclude = ("timestamp", "source")
            cols = [x for x in df.columns if x not in exclude]
            entry["average"].update(df[cols].mean().to_dict())
            entry["minimum"].update(df[cols].min().to_dict())
            entry["maximum"].update(df[cols].max().to_dict())
            stats.append(entry)

        return stats

    def _show_stats(self, show_all_timestamps=True):
        avg_across_batches = pd.DataFrame()
        for batch, df in self._df_by_batch.items():
            if df.empty:
                continue
            if show_all_timestamps:
                print(tabulate(df, headers="keys", tablefmt="psql", showindex=True))
                print(f"Title = {self.metric()} {batch}\n")
                print("\n", end="")
            table = PrettyTable(title=f"{self.metric()} {batch} summary")
            row = ["Average"]
            exclude = ("timestamp", "source")
            cols = [x for x in df.columns if x not in exclude]
            table.field_names = ["stat"] + cols
            average = df[cols].mean()
            avg_across_batches[batch] = average
            for field, val in average.to_dict().items():
                if field not in exclude:
                    row.append(self.get_printable_value(field, val))
            table.add_row(row)
            row = ["Minimum"]
            for field, val in df.min().to_dict().items():
                if field not in exclude:
                    row.append(self.get_printable_value(field, val))
            table.add_row(row)
            row = ["Maximum"]
            for field, val in df.max().to_dict().items():
                if field not in exclude:
                    row.append(self.get_printable_value(field, val))
            table.add_row(row)
            print(table)
            print("\n", end="")

        table = PrettyTable(title=f"{self.metric()} averages per interval across batches")
        averages = avg_across_batches.transpose().mean().to_dict()
        table.field_names = list(averages.keys())
        row = [self.get_printable_value(k, v) for k, v in averages.items()]
        if row:
            table.add_row(row)
            print(table)
        else:
            print("No events are stored")
        print("\n", end="")

    def show_stat_totals(self, stats_to_total):
        """Print a table that shows statistic totals by batch.

        Parameters
        ----------
        stats_to_total : list

        """
        table = PrettyTable(title=f"{self.metric()} Totals")
        table.field_names = ["source"] + list(stats_to_total)
        totals = {}
        for batch, df in self._df_by_batch.items():
            row = [batch]
            for stat in stats_to_total:
                total = df[stat].sum()
                if stat not in totals:
                    totals[stat] = total
                else:
                    totals[stat] += total
                val = self.get_printable_value(stat, total)
                row.append(val)
            table.add_row(row)

        if totals:
            total_row = ["total"]
            for stat, val in totals.items():
                total_row.append(self.get_printable_value(stat, val))
            table.add_row(total_row)
            print(table)


class CpuStatsViewer(StatsViewerBase):
    """Shows CPU statistics"""

    def __init__(self, events):
        super(CpuStatsViewer, self).__init__(events, EVENT_NAME_CPU_STATS)

    @staticmethod
    def metric():
        return "CPU"


class DiskStatsViewer(StatsViewerBase):
    """Shows disk statistics"""

    def __init__(self, events):
        super(DiskStatsViewer, self).__init__(events, EVENT_NAME_DISK_STATS)

    @staticmethod
    def metric():
        return "Disk"

    @staticmethod
    def get_printable_value(field, val):
        if field in ("read_bytes", "write_bytes"):
            val = bytes2human(val)
        elif isinstance(val, float):
            val = "{:.3f}".format(val)
        return val

    def show_stats(self, show_all_timestamps=True):
        print("\nDisk statistics for each batch")
        print("==============================\n")
        self._show_stats(show_all_timestamps=show_all_timestamps)

        stats_to_total = (
            "read_bytes",
            "read_count",
            "write_bytes",
            "write_count",
            "read_time",
            "write_time",
        )
        self.show_stat_totals(stats_to_total)


class MemoryStatsViewer(StatsViewerBase):
    """Shows Memory statistics"""

    def __init__(self, events):
        super(MemoryStatsViewer, self).__init__(events, EVENT_NAME_MEMORY_STATS)

    @staticmethod
    def metric():
        return "Memory"

    @staticmethod
    def get_printable_value(field, val):
        if field == "percent":
            val = "{:.3f}".format(val)
        else:
            val = bytes2human(val)
        return val


class NetworkStatsViewer(StatsViewerBase):
    """Shows Network statistics"""

    def __init__(self, events):
        super(NetworkStatsViewer, self).__init__(events, EVENT_NAME_NETWORK_STATS)

    @staticmethod
    def metric():
        return "Network"

    @staticmethod
    def get_printable_value(field, val):
        if field in ("bytes_recv", "bytes_sent"):
            val = bytes2human(val)
        elif isinstance(val, float):
            val = "{:.3f}".format(val)
        return val

    def show_stats(self, show_all_timestamps=True):
        print("\nNetwork statistics for each batch")
        print("=================================\n")
        if not self._df_by_batch:
            print("No events are stored")
            return

        self._show_stats(show_all_timestamps=show_all_timestamps)
        stats_to_total = (
            "bytes_recv",
            "bytes_sent",
            "dropin",
            "dropout",
            "errin",
            "errout",
            "packets_recv",
            "packets_sent",
        )
        self.show_stat_totals(stats_to_total)


class ProcessStatsViewer(StatsViewerBase):
    """Shows process statistics"""

    def __init__(self, events):
        super().__init__(events, EVENT_NAME_PROCESS_STATS)

    @staticmethod
    def metric():
        return "Process"

    @staticmethod
    def _non_plottable_columns():
        """Return the columns that cannot be plotted."""
        return {"name", "source"}

    def get_stats_summary(self):
        """Return a list of objects describing summaries of all stats.

        Returns
        -------
        list
            list of dicts

        """
        stats = []
        for batch, df in self._df_by_batch.items():
            if df.empty:
                continue
            for name, df_name in df.groupby(by="name"):
                entry = {
                    "type": self.metric(),
                    "name": name,
                    "batch": batch,
                    "average": {},
                    "minimum": {},
                    "maximum": {},
                }
                exclude = ("timestamp", "source", "name")
                cols = [x for x in df_name.columns if x not in exclude]
                entry["average"].update(df_name[cols].mean().to_dict())
                entry["minimum"].update(df_name[cols].min().to_dict())
                entry["maximum"].update(df_name[cols].max().to_dict())
                stats.append(entry)

        return stats

    def plot_to_file(self, output_dir):
        if pd.options.plotting.backend != "plotly":
            pd.options.plotting.backend = "plotly"

        exclude = self._non_plottable_columns()
        figures = {}  # column to go.Figure
        for name in self.iter_batch_names():
            df = self.get_dataframe(name)
            for pname, df_name in df.groupby(by="name"):
                for col in (x for x in df_name.columns if x not in exclude):
                    if col not in figures:
                        figures[col] = go.Figure()
                    series = df_name[col]
                    trace_name = f"{name} {pname}".replace("resource_monitor_", "")
                    figures[col].add_trace(go.Scatter(x=series.index, y=series, name=trace_name))

        for col, fig in figures.items():
            title = f"{self.__class__.__name__} {col}"
            fig.update_layout(title=title)
            filename = Path(output_dir) / f"{self.__class__.__name__}__{col}.html"
            fig.write_html(str(filename))
            logger.info("Generated plot in %s", filename)

    def show_stats(self, show_all_timestamps=True):
        text = f"{self.metric()} statistics for each job"
        print(f"\n{text}")
        print("=" * len(text) + "\n")
        self._show_stats(show_all_timestamps=show_all_timestamps)

    def _show_stats(self, show_all_timestamps=True):
        avg_across_processes = pd.DataFrame()
        for batch, df in self._df_by_batch.items():
            if df.empty:
                continue
            if show_all_timestamps:
                print(tabulate(df, headers="keys", tablefmt="psql", showindex=True))
                print(f"Title = {self.metric()} {batch}\n")
                print("\n", end="")
            for name, df_name in df.groupby(by="name"):
                table = PrettyTable(title=f"{self.metric()} {name} summary")
                row = ["Average"]
                exclude = ("timestamp", "source", "name")
                cols = [x for x in df_name.columns if x not in exclude]
                table.field_names = ["stat"] + cols
                average = df_name[cols].mean()
                avg_across_processes[name] = average
                for field, val in average.to_dict().items():
                    if field not in exclude:
                        row.append(self.get_printable_value(field, val))
                table.add_row(row)
                row = ["Minimum"]
                for field, val in df_name.min().to_dict().items():
                    if field not in exclude:
                        row.append(self.get_printable_value(field, val))
                table.add_row(row)
                row = ["Maximum"]
                for field, val in df_name.max().to_dict().items():
                    if field not in exclude:
                        row.append(self.get_printable_value(field, val))
                table.add_row(row)
                print(table)
                print("\n", end="")

        table = PrettyTable(title=f"{self.metric()} averages per interval across processes")
        averages = avg_across_processes.transpose().mean().to_dict()
        table.field_names = list(averages.keys())
        row = [self.get_printable_value(k, v) for k, v in averages.items()]
        if row:
            table.add_row(row)
            print(table)
        else:
            print("No events are stored")
        print("\n", end="")



"""Models for Singularity containers"""

from pathlib import Path
from typing import Optional

from pydantic.v1 import Field, validator

from jade.models import JadeBaseModel


SINGULARITY_SETUP_COMMANDS = """module load singularity-container
export LD_LIBRARY_PATH=/usr/lib64:/nopt/slurm/current/lib64/slurm:$LD_LIBRARY_PATH
export PATH=$PATH:/nopt/slurm/current/bin
"""

BINDS = "/nopt,/usr/lib64/libreadline.so.6,/usr/lib64/libhistory.so.6,/usr/lib64/libtinfo.so.5,/var/run/munge,/usr/lib64/libmunge.so.2,/usr/lib64/libmunge.so.2.0.0,/run/munge"
SINGULARITY_RUN_COMMAND = f"singularity run -B {BINDS} -B /scratch:/scratch -B /projects:/projects"


class SingularityParams(JadeBaseModel):
    """Defines parameters for using Singularity containers"""

    # This must be first for validation to work.
    enabled: bool = Field(
        title="enabled",
        description="Run all jobs through a Singularity container",
        default=False,
    )
    container: Optional[str] = Field(
        title="container",
        description="Path to Singularity container",
    )
    load_command: str = Field(
        title="load_command",
        description="Command to load the singularity environment. Can be empty.",
        default="module load singularity-container",
    )
    run_command: str = Field(
        title="run_command",
        description="Command to run the Singularity container",
        default=SINGULARITY_RUN_COMMAND,
    )
    setup_commands: str = Field(
        title="setup_commands",
        description="Commands to execute within the Singularity container",
        default=SINGULARITY_SETUP_COMMANDS,
    )

    @validator("container")
    def check_container(cls, container, values):
        if values["enabled"]:
            if container is None:
                raise ValueError("'container' must be set")
            if not Path(container).exists():
                raise ValueError(f"{container} does not exist")
        return container



"""Implement JobConfiguration for generic_command."""

from jade.exceptions import InvalidConfiguration
from jade.jobs.job_configuration import JobConfiguration
from jade.extensions.generic_command.generic_command_inputs import GenericCommandInputs


class GenericCommandConfiguration(JobConfiguration):
    """A class used to configure generic_command jobs."""

    def __init__(self, **kwargs):
        """
        Init GenericCommand class

        Parameters
        ----------
        kwargs, extra arguments

        """
        self._cur_job_id = 1
        super(GenericCommandConfiguration, self).__init__(**kwargs)

    @classmethod
    def auto_config(
        cls,
        inputs,
        cancel_on_blocking_job_failure=False,
        minutes_per_job=None,
        append_job_name=False,
        append_output_dir=False,
        **kwargs,
    ):
        """Create a configuration from all available inputs."""
        if isinstance(inputs, str):
            inputs = GenericCommandInputs(inputs)

        config = GenericCommandConfiguration(**kwargs)
        for job_param in inputs.iter_jobs():
            job_param.cancel_on_blocking_job_failure = cancel_on_blocking_job_failure
            job_param.estimated_run_minutes = minutes_per_job
            job_param.append_job_name = append_job_name
            job_param.append_output_dir = append_output_dir
            config.add_job(job_param)

        return config

    def _serialize(self, data):
        pass

    def add_job(self, job):
        # Overrides JobConfiguration.add_job so that it can add a unique
        # identifier to each job.

        # This will not be true at deserialization.
        if job.job_id is None:
            job.job_id = self._cur_job_id
            self._cur_job_id += 1

        if not job.command:
            raise InvalidConfiguration(f"command cannot be emtpy: job_id={job.job_id}")
        self._jobs.add_job(job)

    def create_from_result(self, job, output_dir):
        return None



import logging
import os
import socket
import time
from pathlib import Path

from filelock import SoftFileLock, Timeout

from jade.exceptions import InvalidConfiguration
from jade.jobs.job_configuration import JobConfiguration
from jade.models import ClusterConfig, Job, JobState, JobStatus, SubmissionGroup
from jade.utils.utils import load_data, dump_data, ExtendedJSONEncoder


logger = logging.getLogger(__name__)


LOCK_TIMEOUT = 300


class Cluster:
    """Represents the state of the nodes running jobs."""

    CLUSTER_CONFIG_FILE = "cluster_config.json"
    JOB_STATUS_FILE = "job_status.json"
    LOCK_FILE = "cluster_config.json.lock"
    CONFIG_VERSION_FILE = "config_version.txt"
    JOB_STATUS_VERSION_FILE = "job_status_version.txt"
    SUBMITTER_GROUP_FILE = "submitter_groups.json"

    def __init__(self, config, job_status=None, lock_timeout=LOCK_TIMEOUT):
        """Internal constructor. Use create() or deserialize()."""
        self._config = config
        self._timeout = lock_timeout
        self._job_status = job_status
        self._config_hash = None
        self._job_status_hash = None
        self._config_file = self.get_config_file(self._config.path)
        self._job_status_file = self.get_job_status_file(self._config.path)
        self._lock_file = self.get_lock_file(self._config.path)
        self._hostname = socket.gethostname()

        # These two files contain versions that are duplicated with the files
        # above. They exist to allow very quick reads to verify updates.
        self._config_version_file = os.path.join(
            config.path,
            self.CONFIG_VERSION_FILE,
        )
        self._job_status_version_file = os.path.join(
            config.path,
            self.JOB_STATUS_VERSION_FILE,
        )

    @classmethod
    def create(cls, path, jade_config: JobConfiguration, pipeline_stage_num=None):
        """Create a new instance of a Cluster. Promotes itself to submitter.

        Parameters
        ----------
        path : str
            Base directory for a JADE submission
        jade_config : JobConfiguration
        pipeline_stage_num : None | int
            True if the config is one stage of a pipeline

        """
        config = ClusterConfig(
            path=path,
            pipeline_stage_num=pipeline_stage_num,
            num_jobs=jade_config.get_num_jobs(),
            submitter=socket.gethostname(),
            submission_groups=jade_config.submission_groups,
            version=0,
        )
        job_status = JobStatus(
            jobs=[
                Job(
                    name=x.name,
                    blocked_by=x.get_blocking_jobs(),
                    cancel_on_blocking_job_failure=x.cancel_on_blocking_job_failure,
                    state=JobState.NOT_SUBMITTED,
                )
                for x in jade_config.iter_jobs()
            ],
            hpc_job_ids=[],
            version=0,
        )
        cluster = cls(config, job_status=job_status)
        cluster._serialize_config_version()
        cluster._serialize_job_status_version()
        cluster.serialize("create")
        cluster.serialize_jobs("create")
        cluster.serialize_submission_groups(Path(path))
        return cluster

    @classmethod
    def deserialize(cls, path, try_promote_to_submitter=False, deserialize_jobs=False):
        """Deserialize an existing Cluster from a file.

        Parameters
        ----------
        path : str
            Base directory for a JADE submission
        try_promote_to_submitter : bool
            Attempt to promote to submitter
        deserialize_jobs : bool
            Deserialize current job status

        Returns
        -------
        tuple
            cluster and a bool indicating whether promotion occurred

        """
        return cls.do_action_under_lock(
            path,
            cls._deserialize,
            path,
            try_promote_to_submitter=try_promote_to_submitter,
            deserialize_jobs=deserialize_jobs,
        )

    def are_all_jobs_complete(self):
        """Return True if all jobs are complete.

        Returns
        -------
        bool

        """
        return self._do_action_under_lock(self._are_all_jobs_complete)

    def all_jobs_submitted(self):
        """Return true if all jobs have been submitted."""
        return self._config.submitted_jobs == self._config.num_jobs

    def am_i_submitter(self):
        """Return True if the current system is the submitter."""
        return self._config.submitter == self._hostname

    def complete_hpc_job_id(self, job_id):
        """Complete an HPC job.

        Parameters
        ----------
        job_id : str

        """
        return self._do_action_under_lock(self._complete_hpc_job_id, job_id, serialize=True)

    @property
    def config(self):
        """Return the ClusterConfig"""
        return self._config

    @property
    def config_file(self):
        """Return the path to the cluster config file.

        Returns
        -------
        str

        """
        return self.get_config_file(self._config.path)

    def delete_files_internal(self):
        # This should not be used in normal circumstances.
        for filename in (
            self.CLUSTER_CONFIG_FILE,
            self.JOB_STATUS_FILE,
            self.CONFIG_VERSION_FILE,
            self.JOB_STATUS_VERSION_FILE,
        ):
            path = os.path.join(self.config.path, filename)
            os.remove(path)
            logger.debug("Deleted %s", path)

    def demote_from_submitter(self, serialize=True):
        """Clear the submitter, which must be the current system."""
        return self._do_action_under_lock(self._demote_from_submitter, serialize=serialize)

    def deserialize_jobs(self):
        """Deserialize the current job status."""
        return self._do_action_under_lock(self._deserialize_jobs, self._config.path)

    @staticmethod
    def do_action_under_lock(path, func, *args, **kwargs):
        """Run a function while holding the lock."""
        lock_file = Cluster.get_lock_file(path)
        return Cluster._do_action_under_lock_internal(lock_file, func, *args, **kwargs)

    @staticmethod
    def get_config_file(path):
        """Return the path to the cluster config file.

        Parameters
        ----------
        path : str
            Base directory for a JADE submission

        """
        return os.path.join(path, Cluster.CLUSTER_CONFIG_FILE)

    @staticmethod
    def get_job_status_file(path):
        """Return the path to the job status file.

        Parameters
        ----------
        path : str
            Base directory for a JADE submission

        """
        return os.path.join(path, Cluster.JOB_STATUS_FILE)

    @staticmethod
    def get_lock_file(path):
        """Return the path to the lock file for the cluster config.

        Parameters
        ----------
        path : str
            Base directory for a JADE submission

        """
        return os.path.join(path, Cluster.LOCK_FILE)

    def get_status_summary(self, include_jobs=False):
        """Return a dict that summarizes current status.

        Parameters
        ----------
        include_jobs : bool
            Whether to include individual job status

        Returns
        -------
        dict

        """
        not_submitted = self._config.num_jobs - self._config.submitted_jobs
        summary = {
            "is_complete": self.is_complete(),
            "is_canceled": self.is_canceled(),
            "num_jobs": self._config.num_jobs,
            "completed_jobs": self._config.completed_jobs,
            "not_submitted_jobs": not_submitted,
        }

        if include_jobs:
            summary["job_status"] = self._job_status.dict()

        return summary

    def has_submitter(self):
        """Return True if the config has a submitter assigned."""
        return self._config.submitter is not None

    def iter_jobs(self, state=None):
        """Yields each Job.

        Parameters
        ----------
        state : JobState
            If not None, only return jobs that match this state.

        Yields
        ------
        Job

        """
        assert self._job_status is not None
        for job in self._job_status.jobs:
            if state is not None and job.state != state:
                continue
            yield job

    def iter_hpc_job_ids(self):
        """Yields each Job.

        Yields
        ------
        str
            HPC job ID

        """
        assert self._job_status is not None
        for job_id in self._job_status.hpc_job_ids:
            yield job_id

    @property
    def job_status(self):
        """Return the JobStatus"""
        return self._job_status

    def is_canceled(self):
        """Return True if the submission is canceled."""
        return self._config.is_canceled

    def is_complete(self):
        """Return True if the submission is complete."""
        return self._config.is_complete

    def mark_canceled(self):
        """Mark the submission as being canceled."""
        return self._do_action_under_lock(self._mark_canceled)

    def mark_complete(self):
        """Mark the submission as being complete."""
        return self._do_action_under_lock(self._mark_complete)

    def prepare_for_resubmission(self, jobs_to_resubmit, updated_blocking_jobs_by_name):
        """Reset the state of the cluster for resubmission of jobs.

        Parameters
        ----------
        jobs_to_resubmit : set
            job names that will be resubmitted
        updated_blocking_jobs_by_name : dict
            contains the blocking jobs for each job to be resubmitted

        """
        # Locking is not required for this function.
        assert self._config.is_complete
        self._config.is_complete = False
        self._config.submitted_jobs = self._config.num_jobs - len(jobs_to_resubmit)
        self._config.completed_jobs = 0

        for job in self.iter_jobs():
            if job.name in jobs_to_resubmit:
                job.state = JobState.NOT_SUBMITTED
                job.blocked_by = updated_blocking_jobs_by_name.get(job.name, set())
            elif job.state == JobState.DONE:
                self._config.completed_jobs += 1

        self._serialize("prepare_for_resubmission")
        self._serialize_jobs("prepare_for_resubmission")
        self.serialize_submission_groups(Path(self._config.path))

    def promote_to_submitter(self, serialize=True):
        """Promote the current system to submitter.

        Returns
        -------
        bool
            Returns True if promotion was successful
        """
        return self._do_action_under_lock(self._promote_to_submitter, serialize=serialize)

    def serialize(self, reason):
        """Serialize the config to a file."""
        self._do_action_under_lock(self._serialize, reason)

    def serialize_jobs(self, reason):
        """Serialize the job status to a file."""
        self._do_action_under_lock(self._serialize_jobs, reason)

    def serialize_submission_groups(self, directory):
        """Serialize the submission groups so that they can be read without
        acquiring a lock.

        Parameters
        ----------
        directory : Path

        """
        path = directory / self.SUBMITTER_GROUP_FILE
        data = [x.dict() for x in self._config.submission_groups]
        dump_data(data, path, cls=ExtendedJSONEncoder)

    @staticmethod
    def deserialize_submission_groups(directory):
        """Return the submission groups being used by the cluster.

        Parameters
        ----------
        directory : Path

        Returns
        -------
        list
            list of SubmissionGroup

        """
        return [SubmissionGroup(**x) for x in load_data(directory / Cluster.SUBMITTER_GROUP_FILE)]

    def update_job_status(
        self,
        submitted_jobs,
        blocked_jobs,
        canceled_jobs,
        completed_job_names,
        hpc_job_ids,
        batch_index,
    ):
        """Update the job status in the config file.

        Parameters
        ----------
        submitted_jobs : list
            list of Job
        blocked_jobs : list
            list of Job
        canceled_jobs : list
            list of Job
        completed_job_names : set
            set of str
        hpc_job_ids : list
            list of str of newly submitted job IDs
        batch_index : int
            next batch index

        """
        self._do_action_under_lock(
            self._update_job_status,
            submitted_jobs,
            blocked_jobs,
            canceled_jobs,
            completed_job_names,
            hpc_job_ids,
            batch_index,
        )

    def _are_all_jobs_complete(self):
        for job in self.iter_jobs():
            if job.state != JobState.DONE:
                assert (
                    self._config.completed_jobs != self._config.num_jobs
                ), f"completed={self._config.completed_jobs}"
                return False

        assert (
            self._config.completed_jobs == self._config.num_jobs
        ), "completed={self._config.completed_jobs}"
        return True

    def _complete_hpc_job_id(self, job_id, serialize=True):
        self._job_status.hpc_job_ids.remove(job_id)
        logger.info("Completed HPC job_id=%s", job_id)
        if serialize:
            self._serialize_jobs("complete_hpc_job_id")

    def _demote_from_submitter(self, serialize=True):
        assert self.am_i_submitter(), self._config.submitter
        self._config.submitter = None
        if serialize:
            self._serialize("demote_from_submitter")

    @classmethod
    def _deserialize(cls, path, try_promote_to_submitter=False, deserialize_jobs=False):
        config_file = cls.get_config_file(path)
        if not os.path.isfile(config_file):
            raise InvalidConfiguration(f"{config_file} does not exist")

        config = ClusterConfig(**load_data(config_file))
        cluster = cls(config)
        promoted = False
        if try_promote_to_submitter:
            promoted = cluster._promote_to_submitter()
        if deserialize_jobs:
            cluster._deserialize_jobs(path)

        return cluster, promoted

    def _deserialize_jobs(self, path):
        data = load_data(self.get_job_status_file(path))
        self._job_status = JobStatus(**data)

    def _do_action_under_lock(self, func, *args, **kwargs):
        return self._do_action_under_lock_internal(
            self._lock_file, func, *args, timeout=self._timeout, **kwargs
        )

    @staticmethod
    def _do_action_under_lock_internal(lock_file, func, *args, timeout=LOCK_TIMEOUT, **kwargs):
        # Using this instead of FileLock because it will be used across nodes
        # on the Lustre filesystem.
        lock_acquisition_seconds = None
        lock = SoftFileLock(lock_file, timeout=timeout)
        try:
            start = time.time()
            lock.acquire(timeout=timeout)
            lock_acquisition_seconds = time.time() - start
        except Timeout:
            # Picked a default value such that this should not trip. If it does
            # trip under normal circumstances then we need to reconsider this.
            logger.error(
                "Failed to acquire file lock %s within %s seconds. hostname=%s",
                lock_file,
                timeout,
                socket.gethostname(),
            )
            raise

        if lock_acquisition_seconds > 5:
            logger.warning(
                "cluster lock acquisition for %s took %s seconds",
                func.__name__,
                lock_acquisition_seconds,
            )

        try:
            val = func(*args, **kwargs)
            lock.release()
            return val
        except Exception as exc:
            lock.release()
            # SoftFileLock always deletes the file, so create it again.
            # There is a small window where this won't work and another node takes over, so
            # callers need to handle the possibility.
            logger.exception(
                "An exception occurred while holding the Cluster lock. The state of the cluster "
                "is unknown."
            )
            try:
                fd = os.open(lock_file, os.O_WRONLY | os.O_CREAT | os.O_EXCL | os.O_TRUNC)
                os.close(fd)
                logger.error("A deadlock will occur.")
            except (IOError, OSError):
                logger.exception(
                    "Unable to cause deadlock. Another node was promoted to submitter"
                )
            raise exc

    def _get_config_version(self):
        with open(self._config_version_file, "r") as f_in:
            return int(f_in.read().strip())

    def _get_job_status_version(self):
        with open(self._job_status_version_file, "r") as f_in:
            return int(f_in.read().strip())

    def _mark_canceled(self):
        self._config.is_canceled = True
        self._serialize("mark_canceled")

    def _mark_complete(self):
        assert not self._config.is_complete
        self._config.is_complete = True
        self._serialize("mark_complete")

    def _promote_to_submitter(self, serialize=True):
        if self.has_submitter():
            return False

        self._config.submitter = self._hostname
        if serialize:
            self._serialize("promote_to_submitter")

        return True

    def _serialize(self, reason):
        current = self._get_config_version()
        if self._config.version != current:
            raise ConfigVersionMismatch(
                f"expected={current} actual={self._config.version} {reason}"
            )

        # Check the hash before the version update.
        if hash(self._config.json()) != self._config_hash:
            self._config.version += 1
            self._serialize_config_version()
            text = self._config.json()
            self._config_hash = hash(text)
            self._serialize_file(self._config.json(), self._config_file)
            logger.info(
                "Wrote config version %s reason=%s hostname=%s",
                self._config.version,
                reason,
                self._hostname,
            )

    def _serialize_jobs(self, reason):
        current = self._get_job_status_version()
        if self._job_status.version != current:
            raise JobStatusVersionMismatch(
                f"expected={current} actual={self._job_status.version} {reason}"
            )

        # Check the hash before the version update.
        if hash(self._job_status.json()) != self._config_hash:
            self._job_status.version += 1
            self._serialize_job_status_version()
            text = self._job_status.json()
            self._serialize_file(text, self._job_status_file)
            self._job_status_hash = hash(text)
            logger.info(
                "Wrote job_status version %s reason=%s hostname=%s",
                self._job_status.version,
                reason,
                self._hostname,
            )

    @staticmethod
    def _serialize_file(text, filename):
        backup = None
        if os.path.exists(filename):
            backup = filename + ".bk"
            os.rename(filename, backup)
        with open(filename, "w") as f_out:
            f_out.write(text + "\n")
        if backup:
            os.remove(backup)

    def _serialize_config_version(self):
        with open(self._config_version_file, "w") as f_out:
            f_out.write(str(self._config.version) + "\n")

    def _serialize_job_status_version(self):
        with open(self._job_status_version_file, "w") as f_out:
            f_out.write(str(self._job_status.version) + "\n")

    def _update_job_status(
        self,
        submitted_jobs,
        blocked_jobs,
        canceled_jobs,
        completed_job_names,
        hpc_job_ids,
        batch_index,
    ):
        self._job_status.hpc_job_ids = hpc_job_ids
        self._job_status.batch_index = batch_index
        status_lookup = {x.name: x for x in self._job_status.jobs}

        processed = set()
        for job in submitted_jobs:
            assert status_lookup[job.name].state != JobState.SUBMITTED, job.name
            status_lookup[job.name].state = JobState.SUBMITTED
            processed.add(job.name)
            self._config.submitted_jobs += 1

        for job in blocked_jobs:
            old = status_lookup[job.name]
            assert old.state == JobState.NOT_SUBMITTED, f"name={job.name} state={old.state}"
            old.blocked_by = job.blocked_by
            processed.add(job.name)

        for _ in canceled_jobs:
            self._config.submitted_jobs += 1

        for name in completed_job_names:
            assert name not in processed, name
            status_lookup[name].state = JobState.DONE
            self._config.completed_jobs += 1

        for job in self.iter_jobs():
            if job.blocked_by and job.state in (JobState.SUBMITTED, JobState.DONE):
                job.blocked_by.clear()

        self._serialize("update_job_status")
        self._serialize_jobs("update_job_status")

        not_submitted = self._config.num_jobs - self._config.submitted_jobs
        logger.info(
            "Updated job status submitted=%s not_submitted=%s completed=%s hostname=%s",
            self._config.submitted_jobs,
            not_submitted,
            self._config.completed_jobs,
            self._hostname,
        )


class ConfigVersionMismatch(Exception):
    """Raised when user tries to perform an update with an old version."""


class JobStatusVersionMismatch(Exception):
    """Raised when user tries to perform an update with an old version."""
